# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import unittest
from collections import defaultdict

from monasca_agent.collector.checks_d import swift_recon


class FakeLogger(object):
    def __init__(self):
        self.log = {'error': [],
                    'warning': []}

    def _write_msg(self, msg, key):
        self.log[key].append(msg)

    def error(self, msg):
        self._write_msg(msg, 'error')

    def warning(self, msg):
        self._write_msg(msg, 'warning')

    def get_loglines(self, key):
        return self.log[key]


class MockSwiftRecon(swift_recon.SwiftRecon):
    def __init__(self):
        super(MockSwiftRecon, self).__init__(
            name='swift_recon',
            init_config={},
            instances=[],
            agent_config={}
        )
        self.log = FakeLogger()
        self.scout_returns = []
        self.reset_gauge()

    def scout_host(self, base_url, recon_type, timeout=5):
        if not self.scout_returns:
            raise swift_recon.SwiftReconException("Mock error")
        if isinstance(self.scout_returns[0], swift_recon.SwiftReconException):
            raise self.scout_returns.pop(0)
        else:
            return self.scout_returns.pop(0)

    def gauge(self, key, value, dimensions, *args, **kwargs):
        self.gauge_called = True
        self.gauge_calls[key].append(value)
        for k, v in dimensions.items():
            self.dimensions[k].add(v)

    def reset_gauge(self):
        self.gauge_called = False
        self.gauge_calls = defaultdict(list)
        self.dimensions = defaultdict(set)


class SwiftReconTest(unittest.TestCase):
    def setUp(self):
        super(SwiftReconTest, self).setUp()
        self.swiftrecon = MockSwiftRecon()

    def test_to_grafana_date(self):
        for item in (0, 1, 5, 10, 10000, 9.999, "34984", '2303.230420'):
            self.assertEqual(float(item) * 1000,
                             swift_recon.to_grafana_date(item))

    def test_build_base_url(self):
        instance = {'hostname': 'a.great.url', 'port': 1234}
        expected = "http://a.great.url:1234/recon/"
        self.assertEqual(self.swiftrecon._build_base_url(instance), expected)

    def test_base_recon(self):
        instance = {'hostname': 'a.great.url', 'port': 1234}
        self.swiftrecon.scout_returns = [
            ("http://a.great.url:1234/recon/", {}, 200), ]

        # When scout is successful we just get the de-jsoned content and
        # a dimensions dict.
        content, dim = self.swiftrecon._base_recon(instance, 'blah')
        self.assertDictEqual(content, {})

        # An error will return None, None
        content, dim = self.swiftrecon._base_recon(instance, 'blah')
        self.assertIsNone(content)
        self.assertIsNone(dim)

    def _setup_speced_mock(self):
        mocked = MockSwiftRecon()
        mocked.log = FakeLogger()
        mocked.umount_check = mock.Mock()
        mocked.disk_usage = mock.Mock()
        mocked.quarantine_check = mock.Mock()
        mocked.driveaudit_check = mock.Mock()
        mocked.async_check = mock.Mock()
        mocked.object_auditor_check = mock.Mock()
        mocked.updater_check = mock.Mock()
        mocked.expirer_check = mock.Mock()
        mocked.auditor_check = mock.Mock()
        mocked.replication_check = mock.Mock()
        return mocked

    def test_check_missing_options(self):
        # missing hostname
        instance = {'server_type': 'object', 'port': 1234}
        self.swiftrecon.check(instance)
        self.assertIn('Missing hostname',
                      self.swiftrecon.log.get_loglines('error'))

        # missing port
        instance = {'server_type': 'object', 'hostname': 'a.great.url'}
        self.swiftrecon.log = FakeLogger()
        self.swiftrecon.check(instance)
        self.assertIn('Missing port',
                      self.swiftrecon.log.get_loglines('error'))

        # Missing server_type
        mocked_swift = self._setup_speced_mock()
        instance = {'hostname': 'a.great.url', 'port': 1234}
        called = ('umount_check', 'disk_usage', 'quarantine_check',
                  'driveaudit_check')
        not_called = ('async_check', 'object_auditor_check', 'updater_check',
                      'expirer_check', 'auditor_check', 'replication_check')
        mocked_swift.check(instance)
        self.assertIn('Missing server_type, so will only attempt '
                      'common checks',
                      mocked_swift.log.get_loglines('warning'))

        # only checks that aren't server_type related will be tested
        for method in called:
            self.assertTrue(getattr(mocked_swift, method).called)
        for method in not_called:
            self.assertFalse(getattr(mocked_swift, method).called)

    def test_checks_for_object_server_type(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}
        mocked_swift = self._setup_speced_mock()
        called = ('async_check', 'object_auditor_check', 'updater_check',
                  'expirer_check', 'replication_check', 'umount_check',
                  'disk_usage', 'quarantine_check', 'driveaudit_check')
        not_called = ('auditor_check', )
        mocked_swift.check(instance)

        for method in called:
            self.assertTrue(getattr(mocked_swift, method).called)
        for method in not_called:
            self.assertFalse(getattr(mocked_swift, method).called)

    def test_checks_for_container_server_type(self):
        instance = {'server_type': 'container', 'hostname': 'awesome.host',
                    'port': 1234}
        mocked_swift = self._setup_speced_mock()
        called = ('updater_check', 'auditor_check', 'replication_check',
                  'umount_check', 'disk_usage', 'quarantine_check',
                  'driveaudit_check')
        not_called = ('async_check', 'object_auditor_check', 'expirer_check')
        mocked_swift.check(instance)

        for method in called:
            self.assertTrue(getattr(mocked_swift, method).called)
        for method in not_called:
            self.assertFalse(getattr(mocked_swift, method).called)

    def test_checks_for_account_server_type(self):
        instance = {'server_type': 'account', 'hostname': 'awesome.host',
                    'port': 1234}
        mocked_swift = self._setup_speced_mock()
        called = ('auditor_check', 'replication_check',
                  'umount_check', 'disk_usage', 'quarantine_check',
                  'driveaudit_check')
        not_called = ('updater_check', 'async_check', 'object_auditor_check',
                      'expirer_check')
        mocked_swift.check(instance)

        for method in called:
            self.assertTrue(getattr(mocked_swift, method).called)
        for method in not_called:
            self.assertFalse(getattr(mocked_swift, method).called)

    def _test_scout_error_no_gauge(self, func, *args):
        # first time we run a check, a SwiftReconException will be thrown
        # so the gauge mock wont have been called.

        self.swiftrecon.reset_gauge()
        self.swiftrecon.scout_returns = [
            swift_recon.SwiftReconException('test')]
        func(*args)
        self.assertFalse(self.swiftrecon.gauge_called)

    def test_umount_check(self):
        instance = {'server_type': 'account', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        self._test_scout_error_no_gauge(self.swiftrecon.umount_check,
                                        instance)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content = [{"device": "sdb1", "mounted": False},
                         {"device": "sdb5", "mounted": False}]
        self.swiftrecon.scout_returns = [
            (expected_url, scout_content, 200)]

        self.swiftrecon.umount_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)
        self.assertIn('swift_recon.unmounted', self.swiftrecon.gauge_calls)
        self.assertEqual(
            self.swiftrecon.gauge_calls['swift_recon.unmounted'][0], 2)

    def test_disk_usage(self):
        instance = {'server_type': 'account', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        self._test_scout_error_no_gauge(self.swiftrecon.disk_usage,
                                        instance)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content = [
            {"device": "sdb1", "avail": "", "mounted": False,
             "used": "", "size": ""},
            {"device": "sdb5", "avail": "500", "mounted": True,
             "used": "200", "size": "700"}]
        self.swiftrecon.scout_returns = [
            (expected_url, scout_content, 200)]

        self.swiftrecon.disk_usage(instance)
        self.assertTrue(self.swiftrecon.gauge_called)

        for dim in ('sdb1', 'sdb5'):
            self.assertIn(dim, self.swiftrecon.dimensions['device'])

        for stat, count in (('mounted', 2), ('size', 1), ('used', 1),
                            ('avail', 1)):
            self.assertIn('swift_recon.disk_usage.{0}'.format(stat),
                          self.swiftrecon.gauge_calls)
            # We only send int values, so there should only be mounted sent
            # more then once.
            self.assertEqual(
                len(self.swiftrecon.gauge_calls[
                    'swift_recon.disk_usage.{0}'.format(stat)]), count)

    def test_quarantine_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        self._test_scout_error_no_gauge(self.swiftrecon.quarantine_check,
                                        instance)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content = {
            "objects": 0, "accounts": 1, "containers": 2,
            "policies": {
                0: {"objects": 5},
                1: {"objects": 4}
            }}
        self.swiftrecon.scout_returns = [
            (expected_url, scout_content, 200),
            (expected_url, {'objects': 1, 'containers': 2,
                            'accounts': 3}, 200)]

        # first we test the result from a Swift 2+ storage node (has
        # storage polices)
        self.swiftrecon.quarantine_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)

        self.assertIn('swift_recon.quarantined', self.swiftrecon.gauge_calls)
        values = self.swiftrecon.gauge_calls['swift_recon.quarantined']
        self.assertEqual(len(values), 4)
        self.assertListEqual(values, [1, 2, 5, 4])
        self.assertSetEqual(self.swiftrecon.dimensions['ring'],
                            {'account', 'container', 'object', 'object-1'})

        # now let's try a pre-storage policy result (swift <2.0)
        self.swiftrecon.reset_gauge()
        self.swiftrecon.quarantine_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)

        self.assertIn('swift_recon.quarantined', self.swiftrecon.gauge_calls)
        values = self.swiftrecon.gauge_calls['swift_recon.quarantined']
        self.assertEqual(len(values), 3)
        self.assertListEqual(values, [3, 2, 1])
        self.assertSetEqual(self.swiftrecon.dimensions['ring'],
                            {'account', 'container', 'object'})

    def test_driveaudit_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        self._test_scout_error_no_gauge(self.swiftrecon.driveaudit_check,
                                        instance)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)

        self.swiftrecon.scout_returns = [
            (expected_url, {"drive_audit_errors": 5}, 200),
            (expected_url, {"drive_audit_errors": None}, 200)]

        self.swiftrecon.driveaudit_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)
        self.assertIn('swift_recon.drive_audit_errors',
                      self.swiftrecon.gauge_calls)
        self.assertListEqual(
            self.swiftrecon.gauge_calls['swift_recon.drive_audit_errors'], [5])

        # If the result it None the gauge wont be called
        self.swiftrecon.reset_gauge()
        self.swiftrecon.driveaudit_check(instance)
        self.assertFalse(self.swiftrecon.gauge_called)

    def test_async_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        self._test_scout_error_no_gauge(self.swiftrecon.async_check,
                                        instance)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)

        self.swiftrecon.scout_returns = [
            (expected_url, {"async_pending": 12}, 200),
            (expected_url, {"async_pending": None}, 200)]

        self.swiftrecon.async_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)
        self.assertIn('swift_recon.object.async_pending',
                      self.swiftrecon.gauge_calls)
        self.assertListEqual(
            self.swiftrecon.gauge_calls['swift_recon.object.async_pending'],
            [12])

        # If the result it None the gauge wont be called
        self.swiftrecon.reset_gauge()
        self.swiftrecon.async_check(instance)
        self.assertFalse(self.swiftrecon.gauge_called)

    def test_object_auditor_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        self._test_scout_error_no_gauge(self.swiftrecon.object_auditor_check,
                                        instance)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content = {
            "object_auditor_stats_ALL": {
                "passes": 5,
                "errors": 1,
                "audit_time": 4,
                "start_time": 1531724606.053309,
                "quarantined": 2,
                "bytes_processed": 11885},
            "object_auditor_stats_ZBF": {
                "passes": 3,
                "errors": 0,
                "audit_time": 0,
                "start_time": 1531724665.303363,
                "quarantined": 0,
                "bytes_processed": 0}}

        self.swiftrecon.scout_returns = [
            (expected_url, scout_content, 200)]

        self.swiftrecon.object_auditor_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)

        prefix = 'swift_recon.object.auditor.{}'
        scout_key = 'object_auditor_stats_{}'
        for frag in ('ALL', 'ZBF'):
            scout_metric = scout_key.format(frag)
            metric = prefix.format(scout_metric)
            for key in scout_content[scout_metric]:
                self.assertIn('{}.{}'.format(metric, key),
                              self.swiftrecon.gauge_calls)
                if key == 'start_time':
                    # there is a hack to make epochs dates in grafana and that
                    # is to time it by 1000
                    scout_content[scout_metric][key] = \
                        scout_content[scout_metric][key] * 1000
                self.assertListEqual(
                    self.swiftrecon.gauge_calls['{}.{}'.format(metric, key)],
                    [scout_content[scout_metric][key]])

    def test_updater_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        for server_type in ('object', 'container'):
            self._test_scout_error_no_gauge(self.swiftrecon.updater_check,
                                            instance, server_type)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)

        self.swiftrecon.scout_returns = [
            (expected_url, {"object_updater_sweep":
                            0.10095596313476562}, 200),
            (expected_url, {"container_updater_sweep":
                            0.11764812469482422}, 200)]

        metric = 'swift_recon.{0}.{0}_updater_sweep'

        self.swiftrecon.updater_check(instance, 'object')
        self.assertTrue(self.swiftrecon.gauge_called)
        self.assertIn(metric.format('object'), self.swiftrecon.gauge_calls)
        self.assertListEqual(
            self.swiftrecon.gauge_calls[metric.format('object')],
            [0.10095596313476562])
        self.swiftrecon.reset_gauge()

        instance = {'server_type': 'container', 'hostname': 'awesome.host',
                    'port': 1234}
        self.swiftrecon.updater_check(instance, 'container')
        self.assertTrue(self.swiftrecon.gauge_called)
        self.assertIn(metric.format('container'), self.swiftrecon.gauge_calls)
        self.assertListEqual(
            self.swiftrecon.gauge_calls[metric.format('container')],
            [0.11764812469482422])

    def test_expirer_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        for server_type in ('object', 'container'):
            self._test_scout_error_no_gauge(self.swiftrecon.updater_check,
                                            instance, server_type)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content = {"object_expiration_pass": 0.021467924118041992,
                         "expired_last_pass": 5}

        self.swiftrecon.scout_returns = [
            (expected_url, scout_content, 200)]

        metric = 'swift_recon.object.expirer.{}'

        self.swiftrecon.expirer_check(instance)
        self.assertTrue(self.swiftrecon.gauge_called)

        for stat in ('object_expiration_pass', 'expired_last_pass'):
            self.assertIn(metric.format(stat), self.swiftrecon.gauge_calls)

        for key in scout_content:
            self.assertListEqual(
                self.swiftrecon.gauge_calls[metric.format(key)],
                [scout_content[key]])

    def test_auditor_check(self):
        instance = {'server_type': 'container', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        for server_type in ('container', 'account'):
            self._test_scout_error_no_gauge(self.swiftrecon.auditor_check,
                                            instance, server_type)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content_cont = {
            "container_audits_passed": 6,
            "container_auditor_pass_completed": 0.015977859497070312,
            "container_audits_since": 1531714368.710222,
            "container_audits_failed": 0}
        scout_content_acc = {
            "account_audits_passed": 2,
            "account_audits_failed": 1,
            "account_auditor_pass_completed": 7.200241088867188,
            "account_audits_since": 1531415933.143866}

        self.swiftrecon.scout_returns = [
            (expected_url, scout_content_cont, 200),
            (expected_url, scout_content_acc, 200)]

        for server_type, content in (('container', scout_content_cont),
                                     ('account', scout_content_acc)):
            instance['server_type'] = server_type
            self.swiftrecon.auditor_check(instance, server_type)
            self.assertTrue(self.swiftrecon.gauge_called)

            for key in content:
                metric = 'swift_recon.{0}.{1}'.format(server_type, key)
                self.assertIn(metric, self.swiftrecon.gauge_calls)
                if key.endswith('_audits_since'):
                    # there is a hack to make epochs dates in grafana and that
                    # is to time it by 1000
                    content[key] = content[key] * 1000
                self.assertListEqual(self.swiftrecon.gauge_calls[metric],
                                     [content[key]])
            self.swiftrecon.reset_gauge()

    def test_replication_check(self):
        instance = {'server_type': 'object', 'hostname': 'awesome.host',
                    'port': 1234}

        # test that the error case doesn't call gauge
        for server_type in ('object', 'container', 'account'):
            self._test_scout_error_no_gauge(self.swiftrecon.replication_check,
                                            instance, server_type)

        # now check the correct case
        expected_url = self.swiftrecon._build_base_url(instance)
        scout_content_obj = {
            "replication_last": 1531724665.483373,
            "replication_stats": {
                "rsync": 0, "success": 1194, "attempted": 647,
                "remove": 0, "suffix_count": 1460, "failure": 1,
                "hashmatch": 1194, "suffix_hash": 0, "suffix_sync": 0},
            "object_replication_last": 1531724665.483373,
            "replication_time": 0.11197113196055095,
            "object_replication_time": 0.11197113196055095}
        scout_content_cont = {
            "replication_last": 1531724675.549912,
            "replication_stats": {
                "no_change": 8, "rsync": 0, "success": 8,
                "deferred": 0, "attempted": 4, "ts_repl": 0,
                "remove": 0, "remote_merge": 0, "diff_capped": 0,
                "failure": 0, "hashmatch": 0, "diff": 0,
                "start": 1531724675.488349, "empty": 0},
            "replication_time": 0.06156301498413086}
        scout_content_acc = {
            "replication_last": 1531724672.644893,
            "replication_stats": {
                "no_change": 0, "rsync": 0, "success": 10, "deferred": 0,
                "attempted": 7, "ts_repl": 0, "remove": 0, "remote_merge": 0,
                "diff_capped": 0, "failure": 5, "hashmatch": 0, "diff": 0,
                "start": 1531724672.639242, "empty": 0},
            "replication_time": 0.005650997161865234}

        self.swiftrecon.scout_returns = [
            (expected_url, scout_content_obj, 200),
            (expected_url, scout_content_cont, 200),
            (expected_url, scout_content_acc, 200)]

        for server_type, content in (('object', scout_content_obj),
                                     ('container', scout_content_cont),
                                     ('account', scout_content_acc)):
            instance['server_type'] = server_type
            self.swiftrecon.replication_check(instance, server_type)
            self.assertTrue(self.swiftrecon.gauge_called)

            for key in ('replication_last', 'replication_time'):
                metric = 'swift_recon.{0}.{1}'.format(server_type, key)
                self.assertIn(metric, self.swiftrecon.gauge_calls)
                if key == 'replication_last':
                    # there is a hack to make epochs dates in grafana and that
                    # is to time it by 1000
                    content[key] = content[key] * 1000
                self.assertListEqual(self.swiftrecon.gauge_calls[metric],
                                     [content[key]])

            # Currently we are only grabbing the following 3 values. As you
            # can see there are much more. Adding more should be done at some
            # point.
            for key in ('attempted', 'failure', 'success'):
                metric = 'swift_recon.{0}.replication.{1}'.format(server_type,
                                                                  key)
                self.assertIn(metric, self.swiftrecon.gauge_calls)
                self.assertListEqual(self.swiftrecon.gauge_calls[metric],
                                     [content['replication_stats'][key]])
            self.swiftrecon.reset_gauge()



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from oslo_utils import importutils

from monasca_setup import agent_config
from monasca_setup import detection
from monasca_setup.detection import utils

LOG = logging.getLogger(__name__)


class InfluxDBRelay(detection.Plugin):
    """Detects influxdb-relay and sets up its monitoring

    Monitored items:

    * process
    * http_check

    """

    PROC_NAME = 'influxdb-relay'
    """Name of the InfluxDB Relay process expected to be found in the system"""
    DEFAULTS = {
        'bind_address': '127.0.0.1',
        'bind_port': 9096
    }
    RELAY_NODE_ARG_NAME = 'influxdb_relay_node'

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """

        proc = utils.find_process_name(self.PROC_NAME)
        process_found = proc is not None

        config_file = self._get_config_file(proc) if process_found else None
        config_file_found = config_file is not None

        dependencies_installed = self.dependencies_installed()

        self.available = (process_found and config_file_found and
                          dependencies_installed)

        if not self.available:
            err_chunks = []
            if not process_found:
                err_chunks.append('\tinfluxdb-relay plugin cannot locate '
                                  '"%s" process.' % self.PROC_NAME)
            elif not config_file_found:
                err_chunks.append('\tinfluxdb-relay plugin cannot locate '
                                  'configuration file.')
            elif not dependencies_installed:
                err_chunks.append('\tinfluxdb-relay plugin requires "toml" '
                                  'to be installed')
            LOG.warning('Plugin for influxdb-relay will not be configured.\n'
                        'Following issue have to be resolved: %s' %
                        '\n'.join(err_chunks))
        else:
            self._config = self._load_config(config_file)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        LOG.info("\tEnabling the influxdb-relay check")
        config = agent_config.Plugins()

        config.merge(self._monitor_process())
        config.merge(self._monitor_endpoint())

        return config

    def _monitor_process(self):
        LOG.info("\tMonitoring the influxdb-relay process")

        dimensions = {}
        if self.args and self.args.get(self.RELAY_NODE_ARG_NAME):
            dimensions.update({self.RELAY_NODE_ARG_NAME: self.args.get(self.RELAY_NODE_ARG_NAME)})

        return detection.watch_process([self.PROC_NAME],
                                       service='influxdb',
                                       component='influxdb-relay',
                                       exact_match=False,
                                       dimensions=dimensions)

    def _monitor_endpoint(self):
        config = agent_config.Plugins()
        http_conf = self._config.get('http', None)
        if isinstance(http_conf, list):
            http_conf = http_conf[0]

        if http_conf:
            host, port = self._explode_bind_address(http_conf)
            listening = utils.find_addrs_listening_on_port(port)
            if listening:
                LOG.info("\tMonitoring the influxdb-relay ping endpoint")

                dimensions = {'service': 'influxdb',
                              'component': 'influxdb-relay'}
                if self.args and self.args.get(self.RELAY_NODE_ARG_NAME):
                    dimensions.update(
                        {self.RELAY_NODE_ARG_NAME: self.args.get(self.RELAY_NODE_ARG_NAME)})

                instance = {
                    'name': 'influxdb-relay',
                    'url': 'http://%s:%d/ping' % (host, port),
                    'dimensions': dimensions
                }

                config['http_check'] = {
                    'init_config': None,
                    'instances': [instance]
                }
            else:
                LOG.warning('\tinfluxdb-relay[http] is enabled but nothing '
                            'could be found listening at %d port. '
                            'It might have happened that process '
                            'was just killed and hence port %d '
                            'was released.', port, port)

        return config

    def dependencies_installed(self):
        return importutils.try_import('toml', False)

    @staticmethod
    def _explode_bind_address(http_conf):
        bind_address = http_conf['bind-addr']
        path, port = bind_address.split(':')

        if not path:
            path = InfluxDBRelay.DEFAULTS['bind_address']
        if not port:
            port = InfluxDBRelay.DEFAULTS['bind_port']

        return path, int(port)

    @staticmethod
    def _load_config(config_file):
        """Loads toml configuration from specified path.

        Method loads configuration from specified `path`
        and parses it with :py:class:`configparser.RawConfigParser`

        """
        try:
            return importutils.import_module('toml').load(config_file)
        except Exception as ex:
            LOG.error('Failed to parse %s', config_file)
            LOG.exception(ex)

    @staticmethod
    def _get_config_file(proc):
        """Tries to retrieve config file location.

        influxdb-relay is launched with ```-config```
        flag in cmdline. That fact is used by this method.
        If, by any mean, that switch is not part of
        :py:method:`psutil.Process.cmdline`, method simply
        fallbacks to None

        :param proc: current process, :py:class:`psutil.Process`
        :type proc: :py:class:`psutil.Process`
        :return: config file path or None
        :rtype: str or None

        """
        cmdline = proc.as_dict(attrs=['cmdline'])['cmdline']
        config_flag = '-config'
        if config_flag in cmdline:
            pos = cmdline.index(config_flag)
            return cmdline[pos + 1]
        LOG.warning('%s switch was not found in influxdb-relay cmdline',
                    config_flag)
        return None



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ntplib

from monasca_agent.collector.checks import AgentCheck

DEFAULT_NTP_VERSION = 3
DEFAULT_TIMEOUT = 1  # in seconds
DEFAULT_HOST = "pool.ntp.org"
DEFAULT_PORT = "ntp"


class NtpCheck(AgentCheck):
    """Uses ntplib to grab a metric for the ntp offset
    """

    def check(self, instance):
        dimensions = self._set_dimensions(None, instance)
        req_args = {
            'host': instance.get('host', DEFAULT_HOST),
            'port': instance.get('port', DEFAULT_PORT),
            'version': int(instance.get('version', DEFAULT_NTP_VERSION)),
            'timeout': float(instance.get('timeout', DEFAULT_TIMEOUT)),
        }
        dimensions['ntp_server'] = req_args['host']
        try:
            ntp_stats = ntplib.NTPClient().request(**req_args)
        except ntplib.NTPException as err:
            self.log.error("Could not connect to NTP Server: %s" % err)
            self.gauge('ntp.connection_status', 1, dimensions=dimensions)
        else:
            ntp_offset = ntp_stats.offset

            # Use the ntp server's timestamp for the time of the result in
            # case the agent host's clock is messed up.
            ntp_ts = ntp_stats.recv_time
            self.gauge('ntp.offset', ntp_offset, timestamp=ntp_ts, dimensions=dimensions)
            self.gauge('ntp.connection_status', 0, timestamp=ntp_ts, dimensions=dimensions)



# Copyright 2019 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import logging
import os
import shutil
import tempfile
from unittest import mock
import unittest

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
import freezegun

from monasca_agent.collector.checks_d import cert_file_check

LOG = logging.getLogger('monasca_agent.collector.checks.check.cert_file_check')


def generate_selfsigned_cert(expired_in):
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=1024,
        backend=default_backend()
    )
    issuer = subject = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, u'foo')
    ])
    now = datetime.datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
        .subject_name(issuer)
        .issuer_name(subject)
        .public_key(key.public_key())
        .serial_number(1)
        .not_valid_before(now)
        .not_valid_after(now + datetime.timedelta(days=expired_in))
        .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)

    return cert_pem.decode('ascii')


class TestCertificateFileCheck(unittest.TestCase):

    def setUp(self):
        super(TestCertificateFileCheck, self).setUp()
        self.cert_file_check_obj = cert_file_check.CertificateFileCheck(
            name='cert_file_check',
            init_config={},
            instances=[],
            agent_config={}
        )

    def test_cert_file_is_none(self):
        with mock.patch.object(LOG, 'warning') as mock_log:
            self.cert_file_check_obj.check({'foo': 'bar'})
            mock_log.assert_called_with(
                'Instance have no "cert_file" configured.')

    def test_unable_to_read_file(self):
        tmp_certdir = tempfile.mkdtemp(prefix='test-cert-file-check-')
        try:
            bogus_cert_file = os.path.join(tmp_certdir, 'foo')
            with mock.patch.object(LOG, 'warning') as mock_log:
                self.cert_file_check_obj.check({'cert_file': bogus_cert_file})
                mock_log.assert_called_with(
                    'Unable to read certificate from %s' % (bogus_cert_file))
        finally:
            shutil.rmtree(tmp_certdir)

    def test_unable_to_load_file(self):
        tmp_certdir = tempfile.mkdtemp(prefix='test-cert-file-check-')
        try:
            bogus_cert_file = os.path.join(tmp_certdir, 'foo')
            # create a non-PEM formatted certificate file
            with open(bogus_cert_file, 'w') as f:
                f.write('foo')

            with mock.patch.object(LOG, 'warning') as mock_log:
                self.cert_file_check_obj.check({'cert_file': bogus_cert_file})
                mock_log.assert_called_with(
                    'Unable to load certificate from %s. Invalid content.' % (
                        bogus_cert_file))
        finally:
            shutil.rmtree(tmp_certdir)

    def test_check(self):
        tmp_certdir = tempfile.mkdtemp(prefix='test-cert-file-check-')
        try:
            cert_file = os.path.join(tmp_certdir, 'foo')
            # create a self-signed cert that expires in 10 days from now
            with open(cert_file, 'w') as f:
                f.write(generate_selfsigned_cert(10))

            with mock.patch.object(cert_file_check.CertificateFileCheck,
                                   'gauge') as mock_gauge:
                self.cert_file_check_obj.check({'cert_file': cert_file})
                args, kwargs = mock_gauge.call_args
                # make sure the plugin correctly detect the given cert will be
                # expiring in equal or less than 10 days from now
                self.assertEqual('cert_file.cert_expire_days', args[0])
                self.assertLessEqual(args[1], 10)
        finally:
            shutil.rmtree(tmp_certdir)

    def test_check_expired_cert(self):
        tmp_certdir = tempfile.mkdtemp(prefix='test-cert-file-check-')
        try:
            cert_file = os.path.join(tmp_certdir, 'foo')
            # create a self-signed cert that has expired 10 days ago
            with open(cert_file, 'w') as f:
                f.write(generate_selfsigned_cert(10))

            now = datetime.datetime.utcnow()
            back_to_the_future = now + datetime.timedelta(days=20)
            with freezegun.freeze_time(back_to_the_future):
                with mock.patch.object(cert_file_check.CertificateFileCheck,
                                       'gauge') as mock_gauge:
                    self.cert_file_check_obj.check({'cert_file': cert_file})
                    args, kwargs = mock_gauge.call_args
                    # make sure the plugin correctly detect that the given
                    # cert has already expired
                    self.assertEqual('cert_file.cert_expire_days', args[0])
                    self.assertLessEqual(args[1], -9)
        finally:
            shutil.rmtree(tmp_certdir)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Redis check tests.
"""
import logging
import os
import unittest
import subprocess
import time
import pprint
import redis

from tests.common import load_check
from nose.plugins.skip import SkipTest
logger = logging.getLogger()

MAX_WAIT = 20
NOAUTH_PORT = 16379
AUTH_PORT = 26379
DEFAULT_PORT = 6379
MISSING_KEY_TOLERANCE = 0.5


class TestRedis(unittest.TestCase):

    def is_travis(self):
        global logger
        logger.info("Running on travis-ci")
        return "TRAVIS" in os.environ

    def wait4(self, p, pattern):
        """Waits until a specific pattern shows up in the stdout
        """
        out = p.stdout
        loop = 0
        while True:
            l = out.readline()
            if l.find(pattern) > -1:
                break
            else:
                time.sleep(0.1)
                loop += 1
                if loop >= MAX_WAIT:
                    break

    def setUp(self):
        raise SkipTest("Requires Redis installed")
        if not self.is_travis():
            self.redis_noauth = subprocess.Popen(
                ["redis-server", "tests/redisnoauth.cfg"], stdout=subprocess.PIPE)
            self.wait4(self.redis_noauth, "The server is now ready to accept connections")
            self.redis_auth = subprocess.Popen(
                ["redis-server", "tests/redisauth.cfg"], stdout=subprocess.PIPE)
            self.wait4(self.redis_auth, "The server is now ready to accept connections")

    def tearDown(self):
        if not self.is_travis():
            self.redis_noauth.terminate()
            self.redis_auth.terminate()

    def test_redis_auth(self):
        # Test connection with password
        if not self.is_travis():
            # correct password
            r = load_check('redisdb', {}, {})
            instance = {
                'host': 'localhost',
                'port': AUTH_PORT,
                'password': 'datadog-is-devops-best-friend'
            }
            r.check(instance)
            metrics = self._sort_metrics(r.get_metrics())
            assert len(metrics) > 0, "No metrics returned"

            # wrong passwords
            instances = [
                {
                    'host': 'localhost',
                    'port': AUTH_PORT,
                    'password': ''
                },
                {
                    'host': 'localhost',
                    'port': AUTH_PORT,
                    'password': 'badpassword'
                }
            ]
            for instance in instances:
                r = load_check('redisdb', {}, {})
                r.check(instance)
                metrics = self._sort_metrics(r.get_metrics())
                assert len(
                    metrics) == 0, "Should have failed with bad password; got %s instead" % metrics

    def test_redis_default(self):
        # Base test, uses the noauth instance
        if self.is_travis():
            port = DEFAULT_PORT
        else:
            port = NOAUTH_PORT

        instance = {
            'host': 'localhost',
            'port': port
        }

        db = redis.Redis(port=port, db=14)  # Datadog's test db
        db.flushdb()
        db.set("key1", "value")
        db.set("key2", "value")
        db.setex("expirekey", "expirevalue", 1000)

        r = load_check('redisdb', {}, {})
        r.check(instance)
        metrics = self._sort_metrics(r.get_metrics())
        assert metrics, "No metrics returned"

        # Assert we have values, timestamps and dimensions for each metric.
        for m in metrics:
            assert isinstance(m[1], int)    # timestamp
            assert isinstance(m[2], (int, float, long))  # value
            dimensions = m[3]["dimensions"]
            expected_dimensions = {"redis_host": "localhost", "redis_port": port}
            for e in expected_dimensions:
                assert e in dimensions

        def assert_key_present(expected, present, tolerance):
            "Assert we have the rest of the keys (with some tolerance for missing keys)"
            e = set(expected)
            p = set(present)
            assert len(e - p) < tolerance * len(e), pprint.pformat((p, e - p))

        # gauges collected?
        remaining_keys = [m[0] for m in metrics]
        expected = r.GAUGE_KEYS.values()
        assert_key_present(expected, remaining_keys, MISSING_KEY_TOLERANCE)

        # Assert that the keys metrics are tagged by db. just check db0, since
        # it's the only one we can guarantee is there.
        db_metrics = self._sort_metrics(
            [m for m in metrics if m[0] in ['redis.keys', 'redis.expires'] and "redis_db:db14" in m[3]["dimensions"]])
        self.assertEqual(2, len(db_metrics))

        self.assertEqual('redis.expires', db_metrics[0][0])
        self.assertEqual(1, db_metrics[0][2])

        self.assertEqual('redis.keys', db_metrics[1][0])
        self.assertEqual(3, db_metrics[1][2])

        # Run one more check and ensure we get total command count
        # and other rates
        time.sleep(5)
        r.check(instance)
        metrics = self._sort_metrics(r.get_metrics())
        keys = [m[0] for m in metrics]
        assert 'redis.net.commands' in keys

    def _sort_metrics(self, metrics):
        def sort_by(m):
            return m[0], m[1], m[3]
        return sorted(metrics, key=sort_by)

if __name__ == "__main__":
    unittest.main()



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Copyright 2017 Fujitsu LIMITED
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from keystoneauth1 import identity
from keystoneauth1 import session
from keystoneclient import discover

from monasca_agent.common import singleton
from monasca_agent import version as ma_version

LOG = logging.getLogger(__name__)


_DEFAULT_SERVICE_TYPE = 'monitoring'
_DEFAULT_ENDPOINT_TYPE = 'public'


def _sanitize_args(data):
    """Removes keys for which value is None.

    :param data: dictionary with data
    :type data: dict
    :return: cleaned data
    :rtype: dict

    """
    return {k: v for k, v in data.items() if v is not None}


def get_session(**kwargs):
    """Creates new keystone session.

    Method uses :py:class:`keystoneauth1.identity.Password`
    abstracting from underlying Keystone version

    This method is capable of creating a session regardless of
    Keystone version (either v2 or v3). However if:

    - using **Keystone v2** following arguments [domain_id, domain_name,
      project_domain_id and project_domain_name] should not be set. Keystone V2
      does not support authentication with domain scope.
    - using **Keystone v2** following arguments are prohibited:
      [user_domain_id, user_domain_name]
    - using **Keystone v3** be careful with the scope of authentication.
      For more details about scopes refer to identity_tokens_ and v3_identity_

    .. _v3_api: https://docs.openstack.org/api-ref/identity/v3/index.html?expanded=token-authe
    ntication-with-scoped-authorization-detail
    .. _identity_tokens: https://docs.openstack.org/admin-guide/identity-tokens.html

    In overall:

    - for **Keystone V2** following arguments are allowed:
      [auth_url, user_id, username, password, trust_id, tenant_name,
      tenant_id, project_name, project_id].
    * for **Keystone V3** following argumenta are allowed:
      [auth_url, user_id, username, password, user_domain_id, user_domain_name,
      trust_id, project_id, project_name, project_domain_id,
      project_domain_name, domain_id, domain_name, tenant_id, tenant_name]

    However, note that project_id and project_name will override tenant_id
    and tenant_name, as in::

        >>> project_id = project_id or tenant_id
        >>> project_name = project_name or tenant_name

    Arguments tenant_id and tenant_name are kept for sake of
    backward compatibility between two versions of Keystone.

    Note:
        Keystone version is resolved on the runtime
        by keystoneauth1 library

    :param string auth_url: URL of keystone service.
    :param string username: Username for authentication.
    :param string password: Password for authentication.
    :param string user_id: User ID for authentication.
    :param string user_domain_id: User's domain ID for authentication
                            (replaced by default_domain_if if set)
    :param string user_domain_name: User's domain name for authentication
                            (replaced by default_domain_name if set)
    :param string project_id: Project ID for authentication
    :param string project_name: Project Name for authentication
    :param string project_domain_id: Project Domain ID for authentication
    :param string project_domain_name: Project Domain Name for authentication
    :param string tenant_id: Tenant ID for authentication
                            (replaced by project_id if set)
    :param string tenant_name: Tenant Name for authentication
                            (replaced by project_name if set)
    :param string domain_id: Domain ID for authentication.
    :param string domain_name: Domain name for authentication
    :param string trust_id: Trust ID for authentication.
    :param string default_domain_id: Default domain ID for authentication.
    :param string default_domain_name: Default domain name for authentication
    :param float keystone_timeout: A timeout to pass to requests. This should be a
                      numerical value indicating some amount (or fraction)
                      of seconds or 0 for no timeout. (optional, defaults
                      to 0)
    :param verify: The verification arguments to pass to requests. These are of
                   the same form as requests expects, so True or False to
                   verify (or not) against system certificates or a path to a
                   bundle or CA certs to check against or None for requests to
                   attempt to locate and use certificates. (optional, defaults
                   to True)
    :param bool reauthenticate: Should reauthenticate if token expires
                        (optional, defaults to True)
    :return: session instance
    :rtype: keystoneauth1.session.Session

    """

    LOG.debug('Initializing keystone session using generic password')

    auth = identity.Password(
        auth_url=kwargs.get('auth_url', None),
        username=kwargs.get('username', None),
        password=kwargs.get('password', None),
        user_id=kwargs.get('user_id', None),
        user_domain_id=kwargs.get('user_domain_id', None),
        user_domain_name=kwargs.get('user_domain_name', None),
        project_id=kwargs.get('project_id', None),
        project_name=kwargs.get('project_name', None),
        project_domain_id=kwargs.get('project_domain_id', None),
        project_domain_name=kwargs.get('project_domain_name', None),
        tenant_id=kwargs.get('tenant_id', None),
        tenant_name=kwargs.get('tenant_name', None),
        domain_id=kwargs.get('domain_id', None),
        domain_name=kwargs.get('domain_name', None),
        trust_id=kwargs.get('trust_id', None),
        default_domain_id=kwargs.get('default_domain_id', None),
        default_domain_name=kwargs.get('default_domain_name', None),
        reauthenticate=kwargs.get('reauthenticate', True)
    )
    sess = session.Session(auth=auth,
                           app_name='monasca-agent',
                           app_version=ma_version.version_string,
                           user_agent='monasca-agent',
                           timeout=kwargs.get('keystone_timeout', None),
                           verify=kwargs.get('verify', True))
    return sess


def get_client(**kwargs):
    """Creates new keystone client.

    Initializes new keystone client.
    Method does not assume what version of keystone is used.
    That responsibility is delegated to
    :py:class:`keystoneauth1.discover.Discover`.
    Version of the keystone will be the newest one available.

    There are two ways to call this method:

    using existing session object (:py:class:`keystoneauth1.session.Session`::

        >>> s = session.Session(**args)
        >>> c = get_client(session=s)

    initializing new keystone client from credentials::

        >>> c = get_client({'username':'mini-mon', 'password':'test', ...})

    :param kwargs: list of arguments passed to method
    :type kwargs: dict
    :return: keystone client instance
    :rtype: Union[keystoneclient.v3.client.Client,
                  keystoneclient.v2_0.client.Client]
    """

    if 'session' not in kwargs:
        LOG.debug('Initializing fresh keystone client')
        sess = get_session(**kwargs)
    else:
        LOG.debug('Initializing keystone client from existing session')
        sess = kwargs.get('session')

    disc = discover.Discover(session=sess)
    LOG.debug('Available keystone versions are %s' % disc.version_data())

    ks = disc.create_client(**kwargs)
    ks.auth_ref = sess.auth.get_auth_ref(session=sess)
    LOG.info('Using keystone version %s', ks.version)

    return ks


def get_args(config):
    """Utility to extract keystone args from agent's config.

    Method retrieves all keystone related settings, from
    agent's configuration, that are actually set.

    :param config: agent's config
    :type config: dict
    :returns: cleaned args
    :rtype: dict

    """
    raw_args = {
        'auth_url': config.get('keystone_url', None),
        'username': config.get('username', None),
        'password': config.get('password', None),
        'user_id': config.get('user_id', None),
        'user_domain_id': config.get('user_domain_id', None),
        'user_domain_name': config.get('user_domain_name', None),
        'project_id': config.get('project_id', None),
        'project_name': config.get('project_name', None),
        'project_domain_name': config.get('project_domain_name', None),
        'project_domain_id': config.get('project_domain_id', None),
        'domain_id': config.get('domain_id', None),
        'domain_name': config.get('domain_name', None),
        'tenant_id': config.get('tenant_id', None),
        'tenant_name': config.get('tenant_name', None),
        'trust_id': config.get('trust_id', None),
        'default_domain_id': config.get('default_domain_id', None),
        'default_domain_name': config.get('default_domain_name', None),
        'url': config.get('url', None),  # hardcoded monasca-api url
        'service_type': config.get('service_type', _DEFAULT_SERVICE_TYPE),
        'endpoint_type': config.get('endpoint_type', _DEFAULT_ENDPOINT_TYPE),
        'region_name': config.get('region_name', None),
        'keystone_timeout': config.get('keystone_timeout', None),
        'verify': False if config.get('insecure') else config.get('ca_file', None),
        'reauthenticate': config.get('reauthenticate', True)
    }
    clean_args = _sanitize_args(raw_args)

    LOG.debug('Removed %d keys that did not present values in configuration',
              len(raw_args) - len(clean_args))

    return clean_args


class Keystone(object, metaclass=singleton.Singleton):

    def __init__(self, config):
        self._config = get_args(config)
        self._keystone_client = None

    def _init_client(self):
        """Get a new keystone client object.

        For more details see:

        - :py:func:`monasca_agent.common.keystone.get_session(**args)`
        - :py:func:`monasca_agent.common.keystone.get_client(**args)`

        Note:
            This method initializes client only once on
            behalf of its own

        :return: keystone client instance
        :rtype: Union[keystoneclient.v3.client.Client,
                      keystoneclient.v2_0.client.Client]

        """

        if self._keystone_client:
            LOG.debug('Keystone client is already initialized')
            return self._keystone_client

        ks = get_client(**self._config)
        self._keystone_client = ks

        return ks

    def get_monasca_url(self):
        """Retrieves monasca endpoint url.

        monasca endpoint url can be retrieved from two locations:

        * agent configuration (value must be present under api.url key)
        * keystone catalog (requires settings api.service_type,
          api.endpoint_type and api.region_name)

        First method tries low-cost approach: checking if url is available
        in configuration file. If not, it moves to querying the keystone
        catalog

        :return: monasca endpoint url
        :rtype: basestring

        """
        if self._config.get('url', None):
            endpoint = self._config.get('url')
            LOG.debug('Using monasca-api url %s from configuration' % endpoint)
        else:
            # NOTE(trebskit) no need to sanitize these values here
            # as we're using already local (clean) copy
            args = {
                'service_type': self._config.get('service_type'),
                'interface': self._config.get('endpoint_type'),
                'region_name': self._config.get('region_name', None)  # that one has no default
            }
            catalog = self._init_client().auth_ref.service_catalog
            endpoint = catalog.url_for(**args)
            LOG.debug('Using monasca-api url %s from catalog[%s]'
                      % (endpoint, args))

        return endpoint

    def get_token(self):
        """Validate token is project scoped and return it if it is

        project_id and auth_token were fetched when keystone client was created

        """
        return self._init_client().auth_token

    def refresh_token(self):
        """Gets a new keystone client object and token
        This method should be called if the token has expired

        """
        self._keystone_client = None
        return self.get_token()

    def get_session(self):
        """Returns session of this client.

        :return: session instance
        :rtype: keystoneauth1.session.Session

        """
        return self._init_client().session



import json
import os
import socket
import urllib

import monasca_agent.collector.checks as checks


def to_grafana_date(item):
    # grafana can handle epoch style dates, but a bit differently
    # they work if you get the standard epoch and muliply it by 1000
    return float(item) * 1000


class SwiftReconException(Exception):
    def __init__(self, msg, errcode=-1):
        self.message = msg
        self.errcode = errcode


class SwiftRecon(checks.AgentCheck):

    def scout_host(self, base_url, recon_type, timeout=5):
        """Perform the actual HTTP request to obtain swift recon telemetry.

        :param base_url: the base url of the host you wish to check. str of the
                        format 'http://127.0.0.1:6200/recon/'
        :param recon_type: the swift recon check to request.
        :returns: tuple of (recon url used, response body, and status)
        """
        url = base_url + recon_type
        try:
            body = urllib.request.urlopen(url, timeout=timeout).read()
            if isinstance(body, bytes):
                body = body.decode('utf8')
            content = json.loads(body)
            self.log.debug("-> %s: %s" % (url, content))
            status = 200
        except urllib.error.HTTPError as err:
            self.log.error("-> %s: %s" % (url, err))
            raise SwiftReconException(err, err.code)
        except (urllib.error.URLError, socket.timeout) as err:
            self.log.error("-> %s: %s" % (url, err))
            raise SwiftReconException(err)
        return url, content, status

    def _build_base_url(self, instance):
        return "http://%(hostname)s:%(port)s/recon/" % instance

    def _base_recon(self, instance, recon_type):
        try:
            url, content, status = self.scout_host(
                self._build_base_url(instance), recon_type,
                instance.get('timeout', 5))

            dimensions = self._set_dimensions({'service': 'swift'}, instance)
            return content, dimensions.copy()
        except SwiftReconException as ex:
            self.log.error('Error running {0}: ({1}) {2}'.format(
                recon_type, ex.errcode, ex.message))
            return None, None

    def async_check(self, instance):
        content, dimensions = self._base_recon(instance, 'async')
        if content is None or content['async_pending'] is None:
            return None

        self.gauge('swift_recon.object.async_pending',
                   content['async_pending'], dimensions)

    def object_auditor_check(self, instance):
        content, dimensions = self._base_recon(instance, 'auditor/object')
        if content is None:
            return None

        for key in ('object_auditor_stats_ALL', 'object_auditor_stats_ZBF'):
            if key not in content:
                continue
            for item in ('audit_time', 'bytes_processed', 'passes', 'errors',
                         'quarantined'):
                if item not in content[key] or content[key][item] is None:
                    continue
                self.gauge(
                    'swift_recon.object.auditor.{0}.{1}'.format(key, item),
                    content[key][item], dimensions)
            if 'start_time' in content[key] and content[key] is not None:
                self.gauge(
                    'swift_recon.object.auditor.{0}.{1}'.format(
                        key, 'start_time'),
                    to_grafana_date(content[key]['start_time']), dimensions)

    def updater_check(self, instance, server_type='object'):
        content, dimensions = self._base_recon(
            instance, 'updater/{0}'.format(server_type))
        stat = '{0}_updater_sweep'.format(server_type)
        if content is None or content[stat] is None:
            return None

        self.gauge('swift_recon.{0}.{1}'.format(server_type, stat),
                   content[stat], dimensions)

    def expirer_check(self, instance):
        content, dimensions = self._base_recon(instance, 'expirer/object')
        if content is None:
            return None

        for stat in ('object_expiration_pass', 'expired_last_pass'):
            if stat not in content or content[stat] is None:
                continue
            data = content[stat]
            self.gauge(
                'swift_recon.object.expirer.{0}'.format(stat),
                data, dimensions)

    def auditor_check(self, instance, server_type='container'):
        content, dimensions = self._base_recon(
            instance, 'auditor/{0}'.format(server_type))
        if content is None:
            return None

        for stat in ('{0}_auditor_pass_completed'.format(server_type),
                     '{0}_audits_failed'.format(server_type),
                     '{0}_audits_passed'.format(server_type)):

            if stat not in content or content[stat] is None:
                continue
            self.gauge('swift_recon.{0}.{1}'.format(server_type, stat),
                       content[stat], dimensions)

        stat = '{0}_audits_since'.format(server_type)
        if stat not in content or content[stat] is None:
            return None
        self.gauge('swift_recon.{0}.{1}'.format(server_type, stat),
                   to_grafana_date(content[stat]), dimensions)

    def replication_check(self, instance, server_type):
        if not server_type:
            return None

        content, dimensions = self._base_recon(
            instance, 'replication/{0}'.format(server_type))
        if content is None:
            return None

        for stat, is_date in (('replication_time', False),
                              ('replication_last', True)):
            if stat not in content or content[stat] is None:
                continue
            if is_date:
                data = to_grafana_date(content[stat])
            else:
                data = content[stat]
            self.gauge('swift_recon.{0}.{1}'.format(server_type, stat),
                       data, dimensions)

        for stat in ('attempted', 'failure', 'success'):

            if stat not in content['replication_stats'] or \
                    content['replication_stats'][stat] is None:
                continue
            self.gauge('swift_recon.{0}.replication.{1}'.format(server_type,
                                                                stat),
                       content['replication_stats'][stat], dimensions)

    def umount_check(self, instance):
        content, dimensions = self._base_recon(instance, 'unmounted')
        if content is None:
            return None

        self.gauge('swift_recon.unmounted', len(content), dimensions)

    def disk_usage(self, instance):
        content, dimensions = self._base_recon(instance, 'diskusage')
        if content is None:
            return None

        for drive in content:
            if not drive.get('device'):
                continue
            dimensions['device'] = drive['device']
            for stat in ('mounted', 'size', 'used', 'avail'):
                if isinstance(drive[stat], str) and \
                        not drive[stat].isdigit():
                    continue
                self.gauge('swift_recon.disk_usage.{0}'.format(stat),
                           drive[stat], dimensions)

    def get_ringmd5(self, instance):
        content, dimensions = self._base_recon(instance, 'ringmd5')
        if content is None:
            return None

        for ring_file, md5 in content.items():
            ring_file = os.path.basename(ring_file)
            if '.' in ring_file:
                ring_file = ring_file.split('.')[0]
            if md5 is None:
                md5 = ''
            self.gauge(
                'swift_recon.md5.{0}'.format(ring_file), md5, dimensions)

    def get_swiftconfmd5(self, instance):
        content, dimensions = self._base_recon(instance, 'swiftconfmd5')
        if content is None:
            return None

        _junk, md5 = content.items()[0]
        if md5 is None:
            md5 = ''
        self.gauge('swift_recon.md5.swift_conf', md5, dimensions)

    def quarantine_check(self, instance):
        content, dimensions = self._base_recon(instance, 'quarantined')
        if content is None:
            return None

        for stat in ('accounts', 'containers'):
            if stat not in content:
                continue
            dimensions['ring'] = stat[:-1]
            self.gauge('swift_recon.quarantined',
                       content[stat], dimensions)

        if 'policies' in content:
            for pol_id in content['policies']:
                ring = 'object' if not pol_id else 'object-{0}'.format(pol_id)
                dimensions['ring'] = ring
                self.gauge('swift_recon.quarantined',
                           content['policies'][pol_id]['objects'], dimensions)
        elif 'objects' in content:
            dimensions['ring'] = 'object'
            self.gauge('swift_recon.quarantined',
                       content['objects'], dimensions)

    def driveaudit_check(self, instance):
        content, dimensions = self._base_recon(instance, 'driveaudit')
        if content is None or content['drive_audit_errors'] is None:
            return None

        self.gauge('swift_recon.drive_audit_errors',
                   content['drive_audit_errors'], dimensions)

    def version_check(self, instance):
        content, dimensions = self._base_recon(instance, 'version')
        if content is None or content['version'] is None:
            return None

        self.gauge(
            'swift_recon.swift_version', content['version'], dimensions)

    def check(self, instance):
        server_type = instance.get('server_type', '')
        if not server_type:
            self.log.warning('Missing server_type, so will only attempt '
                             'common checks')
            server_type = ''
        if not instance.get('hostname'):
            self.log.error('Missing hostname')
            return None
        if not instance.get('port'):
            self.log.error('Missing port')
            return None
        if server_type.upper() not in ('ACCOUNT', 'CONTAINER', 'OBJECT'):
            self.log.warning('server_type name needs to be either account, '
                             'container or object')

        if server_type == 'object':
            self.async_check(instance)
            self.object_auditor_check(instance)
            self.updater_check(instance, server_type)
            self.expirer_check(instance)
        elif server_type == 'container':
            self.auditor_check(instance, server_type)
            self.updater_check(instance, server_type)
        elif server_type == 'account':
            self.auditor_check(instance, server_type)

        if server_type:
            self.replication_check(instance, server_type)
        self.umount_check(instance)
        self.disk_usage(instance)

        # until we can find a way of sending something like an md5, we can
        # run these
        # self.get_ringmd5(instance)
        # self.get_swiftconfmd5(instance)
        self.quarantine_check(instance)
        self.driveaudit_check(instance)

        # Same with the version string.
        # self.version_check(instance)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
################################################################
#
# Copyright (c) 2017 SUSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 or 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
#
################################################################

from monasca_setup import detection


class Magnum(detection.ServicePlugin):
    """Detect Magnum daemons and setup configuration to monitor them."""

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'container-infra',
            'process_names': ['magnum-api', 'magnum-conductor'],
            'service_api_url': 'http://localhost:9511',
            'search_pattern': '.*versions.*',
        }

        super(Magnum, self).__init__(service_params)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from datetime import timedelta, datetime

from monasca_agent.forwarder.transaction import Transaction, TransactionManager
from monasca_agent.forwarder.daemon import MAX_QUEUE_SIZE, THROTTLING_DELAY


class memTransaction(Transaction):

    def __init__(self, size, manager):
        Transaction.__init__(self)
        self._trManager = manager
        self._size = size
        self._flush_count = 0

        self.is_flushable = False

    def flush(self):
        self._flush_count += 1
        if self.is_flushable:
            self._trManager.tr_success(self)
        else:
            self._trManager.tr_error(self)

        self._trManager.flush_next()


class TestTransaction(unittest.TestCase):

    def setUp(self):
        pass

    def testMemoryLimit(self):
        """Test memory limit as well as simple flush"""

        # No throttling, no delay for replay
        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, timedelta(seconds=0), {'dimensions': {}})

        step = 10
        oneTrSize = (MAX_QUEUE_SIZE / step) - 1
        for i in range(step):
            tr = memTransaction(oneTrSize, trManager)
            trManager.append(tr)

        trManager.flush()

        # There should be exactly step transaction in the list, with
        # a flush count of 1
        self.assertEqual(len(trManager._transactions), step)
        for tr in trManager._transactions:
            self.assertEqual(tr._flush_count, 1)

        # Try to add one more
        tr = memTransaction(oneTrSize + 10, trManager)
        trManager.append(tr)

        # At this point, transaction one (the oldest) should have been removed from the list
        self.assertEqual(len(trManager._transactions), step)
        for tr in trManager._transactions:
            self.assertNotEqual(tr._id, 1)

        trManager.flush()
        self.assertEqual(len(trManager._transactions), step)
        # Check and allow transactions to be flushed
        for tr in trManager._transactions:
            tr.is_flushable = True
            # Last transaction has been flushed only once
            if tr._id == step + 1:
                self.assertEqual(tr._flush_count, 1)
            else:
                self.assertEqual(tr._flush_count, 2)

        trManager.flush()
        self.assertEqual(len(trManager._transactions), 0)

    def testThrottling(self):
        """Test throttling while flushing"""

        # No throttling, no delay for replay
        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, THROTTLING_DELAY, {'dimensions': {}})
        trManager._flush_without_ioloop = True  # Use blocking API to emulate tornado ioloop

        # Add 3 transactions, make sure no memory limit is in the way
        oneTrSize = MAX_QUEUE_SIZE / 10
        for i in range(3):
            tr = memTransaction(oneTrSize, trManager)
            trManager.append(tr)

        # Try to flush them, time it
        before = datetime.now()
        trManager.flush()
        after = datetime.now()
        self.assertTrue((after - before) > 3 * THROTTLING_DELAY - timedelta(microseconds=100000),
                        "before = %s after = %s" % (before, after))


if __name__ == '__main__':
    unittest.main()



#!/bin/env python
# (C) Copyright 2015,2016Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Monasca Agent wrapper for Nagios checks.

"""

import hashlib
import os
import pickle
import socket
import subprocess
import time

from monasca_agent.collector.checks.services_checks import ServicesCheck
from monasca_agent.collector.checks.services_checks import Status
import monasca_agent.common.aggregator as aggregator


class WrapNagios(ServicesCheck):

    """Inherit ServicesCheck class to process Nagios checks.

    """

    def __init__(self, name, init_config, agent_config, instances=None):
        ServicesCheck.__init__(self, name, init_config, agent_config, instances)

    @staticmethod
    def _do_skip_check(instance, last_run_data):
        """Determine whether or not to skip a check depending on
        the checks's check_interval, if specified, and the last
        time the check was run
        """
        if instance['name'] in last_run_data and 'check_interval' in instance:
            if time.time() < last_run_data[instance['name']] + instance['check_interval']:
                return True
        else:
            return False

    def _check(self, instance):
        """Run the command specified by check_command and capture the result.

        """

        dimensions = self._set_dimensions({'observer_host': socket.getfqdn()}, instance)

        if 'host_name' in instance:
            dimensions.update({'target_host': instance['host_name']})
        else:
            dimensions.update({'target_host': socket.getfqdn()})

        extra_path = self.init_config.get('check_path')
        env = {}
        env['PATH'] = os.environ['PATH']
        if extra_path:
            env['PATH'] = "{0}:{1}".format(extra_path, env['PATH'])

        last_run_path = self.init_config.get('temp_file_path')
        # Use a default last_run_file if no temp_file is specified in the YAML
        if last_run_path is None:
            last_run_path = '/dev/shm/'  # nosec

        if last_run_path.endswith('/') is False:
            last_run_path += '/'
        last_run_file = (
            last_run_path +
            'nagios_wrapper_' +
            hashlib.md5(
                instance['name']).hexdigest() +
            '.pck')

        # Load last-run data from shared memory file
        last_run_data = {}
        if os.path.isfile(last_run_file):
            file_r = open(last_run_file, "r")
            last_run_data = pickle.load(file_r)
            file_r.close()

        # Exit here if it is not yet time to re-run this check
        if self._do_skip_check(instance, last_run_data) is True:
            return

        metric_name = instance.get('metric_name', instance['name'])

        detail = None
        try:
            proc = subprocess.Popen(instance['check_command'].split(" "),
                                    env=env,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            output = proc.communicate()
            # The check detail is all the text before the pipe
            detail = output[0].split('|')[0]
        except OSError:
            # Return an UNKNOWN code (3) if I have landed here
            error_string = instance['check_command'].split(" ")[0] + " is missing or unreadable"
            self.gauge(metric_name,
                       3,
                       dimensions=dimensions,
                       value_meta={'error': error_string})
            self.log.error(error_string)
            return Status.DOWN, "DOWN: {0}".format(error_string)
        finally:
            # Save last-run data
            last_run_data[instance['name']] = time.time()
            with open(last_run_file, "w") as file_w:
                pickle.dump(last_run_data, file_w)

        status_code = proc.poll()
        if detail:
            value_meta = {'detail': detail}
            overage = aggregator.get_value_meta_overage(value_meta)
            if overage:
                value_meta = {'detail': detail[:-overage]}
            self.gauge(metric_name, status_code,
                       dimensions=dimensions,
                       value_meta=value_meta)
        else:
            self.gauge(metric_name, status_code, dimensions=dimensions)
        # Return DOWN on critical, UP otherwise
        if status_code == "2":
            return Status.DOWN, "DOWN: {0}".format(detail)
        return Status.UP, "UP: {0}".format(detail)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urllib

from monasca_agent.collector.checks import AgentCheck
from monasca_agent.collector.checks.utils import add_basic_auth
from monasca_agent.common.util import headers


class Lighttpd(AgentCheck):

    """Tracks basic connection/requests/workers metrics

    See http://redmine.lighttpd.net/projects/1/wiki/Docs_ModStatus for Lighttpd details
    See http://redmine.lighttpd.net/projects/lighttpd2/wiki/Mod_status for Lighttpd2 details
    """

    URL_SUFFIX_PER_VERSION = {
        1: '?auto',
        2: '?format=plain',
        'Unknown': '?auto'
    }

    GAUGES = {
        'IdleServers': 'lighttpd.performance.idle_server',
        'BusyServers': 'lighttpd.performance.busy_servers',
        'Uptime': 'lighttpd.performance.uptime',
        'Total kBytes': 'lighttpd.net.bytes',
        'Total Accesses': 'lighttpd.net.hits',
        'memory_usage': 'lighttpd.performance.memory_usage',
        'requests_avg': 'lighttpd.net.requests_avg',
        'traffic_out_avg': 'lighttpd.net.bytes_out_avg',
        'traffic_in_avg': 'lighttpd.net.bytes_in_avg',
        'connections_avg': 'lighttpd.net.connections_avg',
        'connection_state_start': 'lighttpd.connections.state_start',
        'connection_state_read_header': 'lighttpd.connections.state_read_header',
        'connection_state_handle_request': 'lighttpd.connections.state_handle_request',
        'connection_state_write_response': 'lighttpd.connections.state_write_response',
        'connection_state_keep_alive': 'lighttpd.connections.state_keep_alive',
        'requests_avg_5sec': 'lighttpd.net.requests_avg_5sec',
        'traffic_out_avg_5sec': 'lighttpd.net.bytes_out_avg_5sec',
        'traffic_in_avg_5sec': 'lighttpd.net.bytes_in_avg_5sec',
        'connections_avg_5sec': 'lighttpd.net.connections_avg_5sec',
    }

    COUNTERS = {
        'requests_abs': 'lighttpd.net.requests_total',
        'traffic_out_abs': 'lighttpd.net.bytes_out',
        'traffic_in_abs': 'lighttpd.net.bytes_in',
        'connections_abs': 'lighttpd.net.connections_total',
        'status_1xx': 'lighttpd.response.status_1xx',
        'status_2xx': 'lighttpd.response.status_2xx',
        'status_3xx': 'lighttpd.response.status_3xx',
        'status_4xx': 'lighttpd.response.status_4xx',
        'status_5xx': 'lighttpd.response.status_5xx',
    }

    RATES = {
        'Total kBytes': 'lighttpd.net.bytes_per_s',
        'Total Accesses': 'lighttpd.net.request_per_s'
    }

    def __init__(self, name, init_config, agent_config, instances=None):
        AgentCheck.__init__(self, name, init_config, agent_config, instances)
        self.assumed_url = {}

    def check(self, instance):
        if 'lighttpd_status_url' not in instance:
            raise Exception("Missing 'lighttpd_status_url' variable in Lighttpd config")

        url = self.assumed_url.get(instance['lighttpd_status_url'], instance['lighttpd_status_url'])

        dimensions = self._set_dimensions(None, instance)
        self.log.debug("Connecting to %s" % url)
        req = urllib.request.Request(url, None, headers(self.agent_config))
        if 'user' in instance and 'password' in instance:
            add_basic_auth(req, instance['user'], instance['password'])
        request = urllib.request.urlopen(req)
        headers_resp = request.info().headers
        server_version = self._get_server_version(headers_resp)
        response = request.read()

        metric_count = 0
        # Loop through and extract the numerical values
        for line in response.split('\n'):
            values = line.split(': ')
            if len(values) == 2:  # match
                metric, value = values
                try:
                    value = float(value)
                except ValueError:
                    continue

                # Special case: kBytes => bytes
                if metric == 'Total kBytes':
                    value *= 1024

                # Send metric as a gauge, if applicable
                if metric in self.GAUGES:
                    metric_count += 1
                    metric_name = self.GAUGES[metric]
                    self.gauge(metric_name, value, dimensions=dimensions)

                # Send metric as a rate, if applicable
                if metric in self.RATES:
                    metric_count += 1
                    metric_name = self.RATES[metric]
                    self.rate(metric_name, value, dimensions=dimensions)

                # Send metric as a counter, if applicable
                if metric in self.COUNTERS:
                    metric_count += 1
                    metric_name = self.COUNTERS[metric]
                    self.increment(metric_name, value, dimensions=dimensions)

        if metric_count == 0:
            url_suffix = self.URL_SUFFIX_PER_VERSION[server_version]
            if self.assumed_url.get(instance['lighttpd_status_url'],
                                    None) is None and url[-len(url_suffix):] != url_suffix:
                self.assumed_url[instance['lighttpd_status_url']] = '%s%s' % (url, url_suffix)
                self.log.warn(
                    "Assuming url was not correct. Trying to add %s suffix to the url" %
                    url_suffix)
                self.check(instance)
            else:
                raise Exception(
                    "No metrics were fetched for this instance. "
                    "Make sure that %s is the proper url." %
                    instance['lighttpd_status_url'])

    def _get_server_version(self, headers):
        for h in headers:
            if "Server:" not in h:
                continue
            try:
                version = int(h.split('/')[1][0])
            except Exception as e:
                self.log.debug("Error while trying to get server version %s" % str(e))
                version = "Unknown"
            self.log.debug("Lighttpd server version is %s" % version)
            return version

        self.log.debug("Lighttpd server version is Unknown")
        return "Unknown"



# (C) Copyright 2015-2017 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from concurrent import futures
import threading

import eventlet
import multiprocessing
from queue import Queue

import monasca_agent.collector.checks

DEFAULT_TIMEOUT = 180
DEFAULT_SIZE_POOL = 6
MAX_LOOP_ITERATIONS = 1000
MAX_ALLOWED_THREADS = 200
FAILURE = "FAILURE"

up_down = collections.namedtuple('up_down', ['UP', 'DOWN'])
Status = up_down('UP', 'DOWN')
EventType = up_down("servicecheck.state_change.up", "servicecheck.state_change.down")


class ServicesCheck(monasca_agent.collector.checks.AgentCheck):
    SOURCE_TYPE_NAME = 'servicecheck'

    """Services checks inherits from this class.

    This class should never be directly instantiated.

    Work flow:
        The main agent loop will call the check function for each instance for
        each iteration of the loop.
        The check method will make an asynchronous call to the _process method in
        one of the thread pool executors created in this class constructor.
        The _process method will call the _check method of the inherited class
        which will perform the actual check.

        The _check method must return a tuple which first element is either
            Status.UP or Status.DOWN.
            The second element is a short error message that will be displayed
            when the service turns down.
    """

    def __init__(self, name, init_config, agent_config, instances):
        monasca_agent.collector.checks.AgentCheck.__init__(
            self, name, init_config, agent_config, instances)

        # A dictionary to keep track of service statuses
        self.statuses = {}
        self.notified = {}
        self.resultsq = Queue()
        self.nb_failures = 0
        self.pool = None

        # The pool size should be the minimum between the number of instances
        # and the DEFAULT_SIZE_POOL. It can also be overridden by the 'threads_count'
        # parameter in the init_config of the check
        try:
            default_size = min(self.instance_count(), multiprocessing.cpu_count() + 1)
        except NotImplementedError:
            default_size = min(self.instance_count(), DEFAULT_SIZE_POOL)
        self.pool_size = int(self.init_config.get('threads_count', default_size))
        self.timeout = int(self.agent_config.get('timeout', DEFAULT_TIMEOUT))

    def start_pool(self):
        if self.pool is None:
            self.log.info("Starting Thread Pool Exceutor")
            self.pool = futures.ThreadPoolExecutor(max_workers=self.pool_size)
            if threading.activeCount() > MAX_ALLOWED_THREADS:
                self.log.error('Thread count (%d) exceeds maximum (%d)' % (threading.activeCount(),
                                                                           MAX_ALLOWED_THREADS))
            self.running_jobs = {}

    def stop_pool(self):
        self.log.info("Stopping Thread Pool")
        if self.pool:
            self.pool.shutdown(wait=True)
            self.pool = None

    def restart_pool(self):
        self.stop_pool()
        self.start_pool()

    def check(self, instance):
        self.start_pool()
        name = instance.get('name', None)
        if name is None:
            self.log.error('Each service check must have a name')
            return

        if (name not in self.running_jobs) or self.running_jobs[name].done():
            # A given instance should be processed one at a time
            self.running_jobs[name] = self.pool.submit(self._process, instance)
        else:
            self.log.info("Instance: %s skipped because it's already running." % name)

    def _process(self, instance):
        name = instance.get('name', None)
        try:
            with eventlet.timeout.Timeout(self.timeout):
                return_value = self._check(instance)
            if not return_value:
                return
            status, msg = return_value
            self._process_result(status, msg, name, instance)
        except eventlet.Timeout:
            msg = 'ServiceCheck {0} timed out'.format(name)
            self.log.error(msg)
            self._process_result(FAILURE, msg, name, instance)
        except Exception:
            msg = 'Failure in ServiceCheck {0}'.format(name)
            self.log.exception(msg)
            self._process_result(FAILURE, msg, name, instance)
        finally:
            del self.running_jobs[name]

    def _process_result(self, status, msg, name, queue_instance):
        if name not in self.statuses:
            self.statuses[name] = []

        self.statuses[name].append(status)

        window = int(queue_instance.get('window', 1))

        if window > 256:
            self.log.warning("Maximum window size (256) exceeded, defaulting it to 256")
            window = 256

        threshold = queue_instance.get('threshold', 1)

        if len(self.statuses[name]) > window:
            self.statuses[name].pop(0)

        nb_failures = self.statuses[name].count(Status.DOWN)

        if nb_failures >= threshold:
            if self.notified.get(name, Status.UP) != Status.DOWN:
                self.notified[name] = Status.DOWN
        else:
            if self.notified.get(name, Status.UP) != Status.UP:
                self.notified[name] = Status.UP

    def _check(self, instance):
        """This function should be implemented by inherited classes.

        """
        raise NotImplementedError



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from unittest import mock
from unittest import TestCase

import monasca_setup
import monasca_setup.agent_config

DEFAULT_HTTP_CHECK_CONFIG = {
    'init_config': None,
    'instances': [{'built_by': 'HttpCheck',
                   'match_pattern': '.*VERSION.*',
                   'url': 'http://127.0.0.1:9200',
                   'name': 'logging',
                   'timeout': '10',
                   'collect_response_time': True,
                   'use_keystone': False,
                   'dimensions': {'service': 'logging'}}]
}

DEFAULT_HTTP_CHECK_2_CONFIG = {
    'init_config': None,
    'instances': [{'built_by': 'HttpCheck',
                   'match_pattern': '.*VERSION.*',
                   'url': 'http://127.0.0.1:9200',
                   'name': 'logging',
                   'timeout': '10',
                   'collect_response_time': True,
                   'use_keystone': False,
                   'dimensions': {'service': 'logging'}},
                  {'built_by': 'HttpCheck',
                   'match_pattern': '.*VERSION.*',
                   'url': 'http://127.0.0.2:9200',
                   'name': 'logging',
                   'timeout': '10',
                   'collect_response_time': True,
                   'use_keystone': False,
                   'dimensions': {'service': 'logging'}}
                  ]
}

DEFAULT_PROCESS_CHECK_CONFIG = {
    'init_config': None,
    'instances': [{'built_by': 'MonNotification',
                   'detailed': True,
                   'dimensions': {'component': 'monasca-notification'},
                   'exact_match': False,
                   'name': 'monasca-notification',
                   'search_string': ['monasca-notification']
                   }]
}

DEFAULT_PROCESS_CHECK_CONFIG_2 = {
    'init_config': None,
    'instances': [{'built_by': 'MonNotification',
                   'detailed': True,
                   'dimensions': {'component': 'monasca-notification'},
                   'exact_match': False,
                   'name': 'monasca-notification',
                   'search_string': ['monasca-notification']
                   },
                  {'built_by': 'MonAPI',
                   'detailed': True,
                   'dimensions': {'component': 'monasca-api'},
                   'exact_match': False,
                   'name': 'monasca-api',
                   'search_string': ['monasca-api']
                   }
                  ]
}

DEFAULT_PING_CHECK_CONFIG_2COMP = {
    'init_config': None,
    'instances': [{'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'dimensions': {'service': 'compute'},
                   'host_name': 'test-test-1-host',
                   'name': 'test-test-1-host ping',
                   'target_hostname': 'test-test-1-mgmt'
                   },
                  {'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'dimensions': {'service': 'compute'},
                   'host_name': 'test-test-2-host',
                   'name': 'test-test-2-host ping',
                   'target_hostname': 'test-test-2-mgmt'
                   },
                  {'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'host_name': 'test-control-1-host',
                   'name': 'test-control-1-host ping',
                   'target_hostname': 'test-control-1-mgmt'
                   }
                  ]
}

DEFAULT_PING_CHECK_CONFIG_ALLCOMP = {
    'init_config': None,
    'instances': [{'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'dimensions': {'service': 'compute'},
                   'host_name': 'test-test-1-host',
                   'name': 'test-test-1-host ping',
                   'target_hostname': 'test-test-1-mgmt'
                   },
                  {'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'dimensions': {'service': 'compute'},
                   'host_name': 'test-test-2-host',
                   'name': 'test-test-2-host ping',
                   'target_hostname': 'test-test-2-mgmt'
                   },
                  {'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'dimensions': {'service': 'compute'},
                   'host_name': 'test-test-3-host',
                   'name': 'test-test-3-host ping',
                   'target_hostname': 'test-test-3-mgmt'
                   },
                  {'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'host_name': 'test-control-1-host',
                   'name': 'test-control-1-host ping',
                   'target_hostname': 'test-control-1-mgmt'
                   }
                  ]
}

DEFAULT_PING_CHECK_CONFIG_CONTONLY = {
    'init_config': None,
    'instances': [{'built_by': 'HostAlive',
                   'alive_test': 'ping',
                   'host_name': 'test-control-1-host',
                   'name': 'test-control-1-host ping',
                   'target_hostname': 'test-control-1-mgmt'
                   }
                  ]
}


INPUT_ARGS = collections.namedtuple(
    "InputArgs", ["overwrite", "user", "config_dir", "detection_args",
                  "detection_plugins", "dry_run", "detection_args_json",
                  "template_dir"])
INPUT_ARGS_WITH_DIMENSIONS = collections.namedtuple(
    "InputArgs", ["overwrite", "user", "config_dir", "detection_args",
                  "detection_plugins", "dry_run", "detection_args_json",
                  "template_dir", "dimensions"])


class TestRemoveConfig(TestCase):
    """ Unit tests for removing_config function in monasca_setup/main.py
        More details are documented in:
        monasca-agent/docs/DeveloperDocs/agent_internals.md
    """
    def setUp(self):
        self._config_data = {}

    def save_config(self, config_dir, plugin_name, user, data):
        self._config_data = data

    # to replace save_plugin_config(args.config_dir, plugin_name, args.user, config)
    #   in delete_config(args, config, file_path, plugin_name)
    def delete_config(self, args, config, file_path, plugin_name):
        self._config_data = config

    @mock.patch('monasca_setup.main.plugin_detection')
    @mock.patch('monasca_setup.main.agent_config.delete_from_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_no_remove_process_check_config(self, mock_read_config,
                                            mock_delete_config,
                                            mock_plugin_detection):
        mock_read_config.return_value = DEFAULT_PROCESS_CHECK_CONFIG
        mock_delete_config.side_effect = self.save_config

        # Add a new process check instance
        same_built_by = DEFAULT_PROCESS_CHECK_CONFIG['instances'][0][
            'built_by']
        same_name = DEFAULT_PROCESS_CHECK_CONFIG['instances'][0][
            'name']
        args, detected_config = self._get_mon_api_check_args_and_config(
            same_built_by, same_name)
        mock_plugin_detection.return_value = detected_config
        self._check_no_change_remove(args, ["HttpCheck"])

    @mock.patch('monasca_setup.main._get_config_yaml_files')
    @mock.patch('monasca_setup.main.plugin_detection')
    @mock.patch('monasca_setup.main.agent_config.delete_from_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_remove_process_check_config(self, mock_read_config,
                                         mock_delete_config,
                                         mock_plugin_detection,
                                         mock_glob):
        mock_read_config.return_value = DEFAULT_PROCESS_CHECK_CONFIG_2
        mock_delete_config.side_effect = self.delete_config
        mock_glob.return_value = ['conf.d--test/process_check-TESTONLY.yaml']

        # Trying to remove mon-api part
        built_by = 'MonAPI'
        name = 'monasca-api'
        args, detected_config = self._get_mon_api_check_args_and_config(
            built_by, name)

        mock_plugin_detection.return_value = detected_config
        print("det_conf {0}".format(detected_config))
        self._check_changes_remove(args, [built_by], DEFAULT_PROCESS_CHECK_CONFIG)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_no_modify_http_check_config(self, mock_read_config,
                                         mock_save_config):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # keep url and match_pattern the same
        same_url = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['url']
        same_match_pattern = DEFAULT_HTTP_CHECK_CONFIG['instances'][0][
            'match_pattern']
        same_name = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['name']

        args, detected_config = self. _get_http_check_args_and_config(
            same_url, same_match_pattern, same_name)
        self._check_no_change_remove(args, detected_config)

    @mock.patch('monasca_setup.main._get_config_yaml_files')
    @mock.patch('monasca_setup.main.plugin_detection')
    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_remove_http_check_config(self, mock_read_config,
                                      mock_save_config,
                                      mock_plugin_detection,
                                      mock_glob):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_2_CONFIG
        mock_save_config.side_effect = self.save_config
        mock_glob.return_value = ['conf.d--test/http_check.yaml']

        # don't change protocol or match_pattern
        url = 'http://127.0.0.2:9200'
        match_pattern = '.*VERSION.*'

        args, detected_config = self. _get_http_check_args_and_config(
            url, match_pattern, 'logging')
        mock_plugin_detection.return_value = detected_config
        expected_value = DEFAULT_HTTP_CHECK_CONFIG
        self._check_changes_remove(args, ['http_check'], expected_value)

    # TODO: test_remove_matching for http or process,
    # TODO: test if no match and test if multiple entries match

    # TODO: start from DEFAULT_PING_CHECK_CONFIG_ALLCOMP and remove the computes by dimension server:compute
    # TODO: do a ping check using JSON format detection arguments, (after that is implemented)

    # start from DEFAULT_PING_CHECK_CONFIG_ALLCOMP and remove just test3 compute
    @mock.patch('monasca_setup.main._get_config_yaml_files')
    @mock.patch('monasca_setup.main.plugin_detection')
    @mock.patch('monasca_setup.main.agent_config.delete_from_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_remove_matching_ping_check_config(self, mock_read_config,
                                               mock_delete_config,
                                               mock_plugin_detection,
                                               mock_glob):
        mock_read_config.return_value = DEFAULT_PING_CHECK_CONFIG_ALLCOMP
        mock_delete_config.side_effect = self.delete_config
        mock_glob.return_value = ['conf.d--test/ping_check.yaml']

        same_built_by = DEFAULT_PING_CHECK_CONFIG_ALLCOMP['instances'][0][
            'built_by']
        # note: NOT testing the json args path
        test_args = INPUT_ARGS(False, 'mon-agent', '/etc/monasca/agent',
                               'hostname=deletehost-localcloud-mgmt '
                               'type=ping '
                               'dimensions=service:compute',
                               [same_built_by], False,
                               {},
                               '/etc/monasca/agent/conf.d--test')

        mock_plugin_detection.return_value = {'ping': {
            'instances': [{'built_by': same_built_by,
                           'alive_test': 'ping',
                           'dimensions': {'service': 'compute'},
                           'host_name': 'test-test-3-host',
                           'name': 'test-test-3-host ping',
                           'target_hostname': 'test-test-3-mgmt'}],
            'init_config': None}
        }

        changes = monasca_setup.main.remove_config_for_matching_args(test_args, [same_built_by])
        self.assertEqual(changes, True, "Should have removed config item but did not!")
        self.assertEqual(DEFAULT_PING_CHECK_CONFIG_2COMP, self._config_data,
                         "Expected value {0} did not match result of {1}"
                         .format(DEFAULT_PING_CHECK_CONFIG_2COMP, self._config_data))

    ####
    # helper functions
    def _check_no_change_remove(self, args, plugins):
        changes = monasca_setup.main.remove_config(args, plugins)
        self.assertEqual(changes, False)
        self.assertEqual(self._config_data, {})

    def _check_changes_remove(self, args, plugins, expected_value):
        changes = monasca_setup.main.remove_config(args, plugins)
        self.assertEqual(changes, True)
        self.assertEqual(expected_value, self._config_data,
                         "Expected value {0} did not match resulting config data {1}"
                         .format(expected_value, self._config_data))

    def _check_changes_remove_matching(self, args, plugins, expected_value):
        changes = monasca_setup.main.remove_config_for_matching_args(args, plugins)
        self.assertEqual(changes, True)
        self.assertEqual(expected_value, self._config_data,
                         "Expected value {0} did not match resulting config data {1}"
                         .format(expected_value, self._config_data))

    # Reminder of args: ["overwrite", "user", "config_dir", "detection_args",
    #                   "detection_plugins", "dry_run",
    #                   "detection_args_json", "template_dir"]
    def _get_mon_api_check_args_and_config(self, built_by, name):
        args = INPUT_ARGS(False, 'mon-agent', '/etc/monasca/agent', None,
                          ['MonAPI'], False,
                          '{}', '/etc/monasca/agent/conf.d--test')
        detected_config = {
            'process':
                {'instances': [{'built_by': built_by,
                                'detailed': True,
                                'dimensions':  {
                                    'component': name},
                                'exact_match': False,
                                'name': name,
                                'search_string': [name]
                                }],
                 'init_config': None
                 }
        }
        return args, detected_config

    def _get_http_check_args_and_config(self, url, match_pattern, name):
        args = INPUT_ARGS(False, 'mon-agent', '/etc/monasca/agent',
                          'url={0} match_pattern={1} name={2} timeout=10 '
                          'use_keystone=False'.format(
                              url, match_pattern, name),
                          ['HttpCheck'], False,
                          '{}', '/etc/monasca/agent/conf.d--test')
        detected_config = {
            'http_check':
                {'instances': [{'built_by': 'HttpCheck',
                                'match_pattern': match_pattern,
                                'url': url,
                                'name': name,
                                'timeout': '10',
                                'collect_response_time': True,
                                'use_keystone': False,
                                'dimensions': {'service': 'logging'}
                                }],
                 'init_config': None
                 }
        }
        return args, detected_config



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import urllib

from monasca_agent.collector.checks import AgentCheck
from monasca_agent.common.util import headers


class CouchDb(AgentCheck):

    """Extracts stats from CouchDB via its REST API

    http://wiki.apache.org/couchdb/Runtime_Statistics
    """

    def _create_metric(self, data, dimensions=None):
        overall_stats = data.get('stats', {})
        for key, stats in overall_stats.items():
            for metric, val in stats.items():
                if val['current'] is not None:
                    metric_name = '.'.join(['couchdb', key, metric])
                    self.gauge(metric_name, val['current'], dimensions=dimensions)

        for db_name, db_stats in data.get('databases', {}).items():
            for name, val in db_stats.items():
                if name in ['doc_count', 'disk_size'] and val is not None:
                    metric_name = '.'.join(['couchdb', 'by_db', name])
                    metric_dimensions = dimensions.copy()
                    metric_dimensions['db'] = db_name
                    self.gauge(metric_name, val, dimensions=metric_dimensions, device_name=db_name)

    def _get_stats(self, url):
        """Hit a given URL and return the parsed json.

        """
        self.log.debug('Fetching Couchdb stats at url: %s' % url)
        req = urllib.request.Request(url, None, headers(self.agent_config))

        # Do the request, log any errors
        request = urllib.request.urlopen(req)
        response = request.read()
        return json.loads(response)

    def check(self, instance):
        server = instance.get('server', None)
        if server is None:
            raise Exception("A server must be specified")
        # Get dimensions
        dimensions = self._set_dimensions({'instance': server}, instance)
        data = self.get_data(server)
        self._create_metric(data, dimensions=dimensions)

    def get_data(self, server):
        # The dictionary to be returned.
        couchdb = {'stats': None, 'databases': {}}

        # First, get overall statistics.
        endpoint = '/_stats/'

        url = '%s%s' % (server, endpoint)
        overall_stats = self._get_stats(url)

        # No overall stats? bail out now
        if overall_stats is None:
            raise Exception("No stats could be retrieved from %s" % url)

        couchdb['stats'] = overall_stats

        # Next, get all database names.
        endpoint = '/_all_dbs/'

        url = '%s%s' % (server, endpoint)
        databases = self._get_stats(url)

        if databases is not None:
            for dbName in databases:
                endpoint = '/%s/' % dbName

                url = '%s%s' % (server, endpoint)
                db_stats = self._get_stats(url)
                if db_stats is not None:
                    couchdb['databases'][dbName] = db_stats

        return couchdb



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import urllib

from monasca_setup import agent_config
from monasca_setup.detection.plugin import Plugin
from monasca_setup.detection.utils import find_addrs_listening_on_port
from monasca_setup.detection.utils import find_process_cmdline
from monasca_setup.detection.utils import service_api_check
from monasca_setup.detection.utils import watch_directory
from monasca_setup.detection.utils import watch_file_size
from monasca_setup.detection.utils import watch_process
from monasca_setup.detection.utils import watch_process_by_username


log = logging.getLogger(__name__)


class ServicePlugin(Plugin):
    """Base class implemented by the monasca-agent plugin detection classes for OpenStack Services.
    Detection plugins inheriting from this class can easily setup up processes to be watched and
    a http endpoint to be checked.

    This class covers Process, HTTP endpoints, Directory, and File monitoring.  It is primarily
    used for monitoring OpenStack components.
    Note: There are existing default detection plugins for http_check.py, directory.py, and
    file_size.py that only require configuration.

    A process can be monitored by process_names or by process_username. Pass in the process_names
    list argument when watching process by name.  Pass in the process_username argument and
    component_name arguments when watching process by username. Watching by username is useful for
    groups of processes that are owned by a specific user.
    For process monitoring by process_username the component_name is required since it is used to
    initialize the instance name in process.yaml.  component_name is optional for monitoring by
    process_name and all other checks.

    An http endpoint connection can be checked by passing in the service_api_url and optional
    search_pattern parameters.
    The http check can be skipped by specifying the argument 'disable_http_check'

    Directory size can be checked by passing in a directory_names list.

    File size can be checked by passing in a file_dirs_names list where each directory name item
    includes a list of files.
    example: 'file_dirs_names': [('/var/log/monasca/api', ['monasca-api'])]

    Note: service_name and component_name are optional (except component_name is required with
    process_username) arguments used for metric dimensions by all checks.
    """

    def __init__(self, kwargs):
        self.service_name = kwargs['service_name']
        self.process_names = kwargs.get('process_names')
        self.process_username = kwargs.get('process_username', None)
        self.component_name = kwargs.get('component_name', None)
        self.file_dirs_names = kwargs.get('file_dirs_names')
        self.directory_names = kwargs.get('directory_names')
        self.service_api_url = kwargs.get('service_api_url')
        self.search_pattern = kwargs.get('search_pattern')
        # overwrite is currently not used with this class, make optional until we remove it.
        overwrite = kwargs.get('overwrite', False)
        template_dir = kwargs['template_dir'],
        if 'args' in kwargs:
            args = kwargs['args']
            if isinstance(args, str):
                try:
                    # Turn 'service_api_url=url' into
                    # dict {'service_api_url':'url'}
                    args_dict = dict(
                        [item.split('=') for item in args.split()])
                    # Allow args to override all of these parameters
                    if 'process_names' in args_dict:
                        self.process_names = args_dict['process_names'].split(',')
                    if 'process_username' in args_dict:
                        self.process_username = args_dict['process_username']
                    if 'component_name' in args_dict:
                        self.component_name = args_dict['component_name']
                    if 'file_dirs_names' in args_dict:
                        self.file_dirs_names = args_dict['file_dirs_names']
                    if 'directory_names' in args_dict:
                        self.directory_names = args_dict['directory_names'].split(',')
                    if 'service_api_url' in args_dict:
                        self.service_api_url = args_dict['service_api_url']
                    if 'search_pattern' in args_dict:
                        self.search_pattern = args_dict['search_pattern']
                    if 'overwrite' in args_dict:
                        overwrite = args_dict['overwrite']
                    if 'template_dir' in args_dict:
                        template_dir = args_dict['template_dir']
                except Exception:
                    log.exception('Error parsing detection arguments')

        super(ServicePlugin, self).__init__(template_dir, overwrite, kwargs.get('args'))

    def _detect(self):
        """Run detection.

        """
        self.found_processes = []
        if self.process_names:
            for process in self.process_names:
                if find_process_cmdline(process) is not None:
                    self.found_processes.append(process)
            if len(self.found_processes) > 0:
                self.available = True
        if self.process_username and self.component_name:
            self.available = True
        if self.file_dirs_names:
            self.available = True
        if self.directory_names:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = agent_config.Plugins()
        if self.found_processes:
            log.info("\tMonitoring by process_name(s): {0} "
                     "for service: {1}.".format(",".join(self.found_processes), self.service_name))
            for process in self.found_processes:
                # Watch the service processes
                component_name = self.component_name if self.component_name else process
                config.merge(watch_process(search_strings=[process], service=self.service_name,
                                           component=component_name, exact_match=False))

        if self.process_username:
            log.info("\tMonitoring by process_username: {0} for "
                     "service: {1}.".format(self.process_username, self.service_name))
            config.merge(watch_process_by_username(username=self.process_username,
                                                   process_name=self.component_name,
                                                   service=self.service_name,
                                                   component=self.component_name))
        if self.file_dirs_names:
            for file_dir_name in self.file_dirs_names:
                # Watch file size
                file_dir = file_dir_name[0]
                file_names = file_dir_name[1]
                if len(file_dir_name) == 3:
                    file_recursive = file_dir_name[2]
                else:
                    file_recursive = False
                if file_names == ['*']:
                    log.info("\tMonitoring the size of all the files in the "
                             "directory {0}.".format(file_dir))
                else:
                    log.info(
                        "\tMonitoring the size of files {0} in the "
                        "directory {1}.".format(
                            ", ".join(
                                str(name) for name in file_names),
                            file_dir))
                config.merge(
                    watch_file_size(
                        directory_name=file_dir,
                        file_names=file_names,
                        file_recursive=file_recursive,
                        service=self.service_name,
                        component=self.component_name))

        if self.directory_names:
            for dir_name in self.directory_names:
                log.info("\tMonitoring the size of directory {0}.".format(
                    dir_name))
                config.merge(
                    watch_directory(
                        directory_name=dir_name,
                        service=self.service_name,
                        component=self.component_name))

        # Skip the http_check if disable_http_check is set
        if self.args is not None and self.args.get('disable_http_check', False):
            self.service_api_url = None
            self.search_pattern = None

        if self.service_api_url and self.search_pattern:
            # Check if there is something listening on the host/port
            parsed = urllib.parse.urlparse(self.service_api_url)
            host, port = parsed.netloc.split(':')
            listening = find_addrs_listening_on_port(port)

            if len(listening) > 0:
                # If not listening on localhost or ips then use another local ip
                if host == 'localhost' and len(
                        set(['127.0.0.1', '0.0.0.0', '::', '::1']) & set(listening)) == 0:
                    new_url = list(parsed)
                    new_url[1] = listening[0] + ':' + port
                    api_url = urllib.parse.urlunparse(new_url)
                else:
                    api_url = self.service_api_url

                # Setup an active http_status check on the API
                log.info("\tConfiguring an http_check for the {0} API.".format(self.service_name))
                config.merge(service_api_check(name=self.service_name + '-api',
                                               url=api_url,
                                               pattern=self.search_pattern,
                                               use_keystone=True,
                                               service=self.service_name,
                                               component=self.component_name))
            else:
                log.info("\tNo process found listening on {0} ".format(port) +
                         "skipping setup of http_check for the {0} API." .format(self.service_name))

        return config

    def dependencies_installed(self):
        """Return True if dependencies are installed.

        """
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import time
import unittest

import monasca_agent.common.config as configuration
from monasca_agent.collector.checks import AgentCheck
from monasca_agent.collector.checks.utils import DynamicCheckHelper

base_config = configuration.Config(os.path.join(os.path.dirname(__file__),
                                                'test-agent.yaml'))


class TestDynamicCheckHelper(unittest.TestCase):
    def setUp(self):
        agent_config = base_config.get_config(sections='Main')

        self._instances = [{'name': 'test',
                            'mapping': {
                                'gauges': ['stats.(MessagesAvg)'],
                                'counters': ['MessagesTotal'],
                                'dimensions': {
                                    'index': 'index',
                                    'simple_dimension': 'simple_label',
                                    'complex_dimension': {
                                        'source_key': 'complex_label',
                                        'regex': 'k8s_([._\-a-zA-Z0-9]*)_postfix'
                                    },
                                    'complex_dimension_rest': {
                                        'source_key': 'complex_label',
                                        'regex': 'k8s_([._\-a-zA-Z0-9]*_postfix)'
                                    }
                                },
                                'groups': {
                                    'testgroup': {
                                        'dimensions': {
                                            'user': 'user'
                                        },
                                        'rates': ['.*\.Responses.*', '(sec_auth_.*).stats',
                                                  '(io_service_bytes)_stats_Total']
                                    }
                                    # dimensions should be inherited from above
                                }}}]
        self.check = AgentCheck("DynCheckHelper-Teset", {}, agent_config, self._instances)  # TODO mock check
        self.helper = DynamicCheckHelper(self.check, 'dynhelper')

    def run_check(self):
        self.check.run()
        metric_dict = {"sec": {"auth": [{"user": "me", "total.stats": 10}, {"user": "you", "total.stats": 15}]},
                       "io_service_bytes": {"stats": {"Total": 10}}}
        self.helper.push_metric_dict(self._instances[0], metric_dict, group="testgroup",
                                     labels={'simple_label': 'simple_label_test',
                                             'complex_label': 'k8s_monasca-api-a8109321_postfix'}, max_depth=3)
        self.helper.push_metric(self._instances[0], metric='req.ResponsesOk', value=10.0,
                                group="testgroup",
                                labels={'simple_label': 'simple_label_test',
                                        'complex_label': 'k8s_monasca-api-a8109321_postfix'})
        self.helper.push_metric(self._instances[0], metric='stats.MessagesAvg', value=5.0,
                                labels={'simple_label': 'simple_label_test',
                                        'complex_label': 'k8s_monasca-api-a8109321_postfix'})
        self.helper.push_metric(self._instances[0], metric='MessagesTotal', value=1)
        time.sleep(1)
        self.helper.push_metric_dict(self._instances[0], metric_dict, group="testgroup",
                                     labels={'simple_label': 'simple_label_test',
                                             'complex_label': 'k8s_monasca-api-a8109321_postfix'}, max_depth=3)
        self.helper.push_metric(self._instances[0], metric='req.ResponsesOk', value=15.0,
                                group="testgroup",
                                labels={'simple_label': 'simple_label_test',
                                        'complex_label': 'k8s_monasca-api-a8109321_postfix'})
        self.helper.push_metric(self._instances[0], metric='MessagesTotal', value=100)
        metrics = self.check.get_metrics()

        return metrics

    def testMeasurements(self):
        metrics = self.run_check()
        for m in metrics:
            print("metric: {0}, dimensions: {1}".format(m['measurement']['name'], repr(m['measurement']['dimensions'])))

        def sortfilter(name):
            filterd = filter(lambda m: m['measurement']['name'] == name, metrics)
            return sorted(filterd, key=lambda m:m['measurement']['timestamp'])

        metric1 = sortfilter('dynhelper.messages_avg')
        metric2 = sortfilter('dynhelper.messages_total')
        metric3 = sortfilter('dynhelper.testgroup.req_responses_ok')
        metric4 = sortfilter('dynhelper.testgroup.sec_auth_total')
        self.assertTrue(len(metric1) > 0,
                        'gauge dynhelper.messages_avg missing in metric list {0}'.format(repr(metrics)))
        self.assertEqual(metric1[0]['measurement']['dimensions'],
                          {'simple_dimension': 'simple_label_test', 'complex_dimension': 'monasca-api-a8109321',
                           'complex_dimension_rest': 'monasca-api-a8109321_postfix',
                           'hostname': metric1[0]['measurement']['dimensions'].get('hostname')})
        self.assertTrue(len(metric2) > 0,
                        'rate dynhelper.messages_total missing in metric list {0}'.format(repr(metrics)))
        self.assertEqual(metric2[0]['measurement']['dimensions'],
                          {'hostname': metric2[0]['measurement']['dimensions'].get('hostname')})
        self.assertTrue(len(metric3) > 0,
                        'rate dynhelper.testgroup.req_responses_ok missing in metric list {0}'.format(repr(metrics)))
        self.assertEqual(metric3[0]['measurement']['dimensions'],
                          {'simple_dimension': 'simple_label_test', 'complex_dimension': 'monasca-api-a8109321',
                           'complex_dimension_rest': 'monasca-api-a8109321_postfix',
                           'hostname': metric3[0]['measurement']['dimensions'].get('hostname')})
        self.assertTrue(len(metric4) == 2,
                        'rate dynhelper.testgroup.sec_auth_total missing in metric list {0}'.format(repr(metrics)))
        self.assertEqual(metric4[0]['measurement']['dimensions'],
                          {'simple_dimension': 'simple_label_test', 'complex_dimension': 'monasca-api-a8109321',
                           'complex_dimension_rest': 'monasca-api-a8109321_postfix',
                           'user': 'me', 'hostname': metric4[0]['measurement']['dimensions'].get('hostname')})
        self.assertEqual(metric4[1]['measurement']['dimensions'],
                          {'simple_dimension': 'simple_label_test', 'complex_dimension': 'monasca-api-a8109321',
                           'complex_dimension_rest': 'monasca-api-a8109321_postfix',
                           'user': 'you', 'hostname': metric4[1]['measurement']['dimensions'].get('hostname')})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import logging
from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest

from tests.common import get_check

logging.basicConfig()

"""
Run the following on your local SQL Server:

CREATE LOGIN datadog WITH PASSWORD = '340$Uuxwp7Mcxo7Khy';
CREATE USER datadog FOR LOGIN datadog;
GRANT SELECT on sys.dm_os_performance_counters to datadog;
GRANT VIEW SERVER STATE to datadog;
"""

CONFIG = """
init_config:
    custom_metrics:
        -   name: sqlserver.clr.execution
            type: gauge
            counter_name: CLR Execution

        -   name: sqlserver.exec.in_progress
            type: gauge
            counter_name: OLEDB calls
            instance_name: Cumulative execution time (ms) per second

        -   name: sqlserver.db.commit_table_entries
            type: gauge
            counter_name: Log Flushes/sec
            instance_name: ALL
            tag_by: db

instances:
    -   host: 127.0.0.1,1433
        username: datadog
        password: 340$Uuxwp7Mcxo7Khy
"""


class SQLServerTestCase(unittest.TestCase):

    @attr('windows')
    def testSqlServer(self):
        raise SkipTest('Requires adodbapi')
        check, instances = get_check('sqlserver', CONFIG)
        check.check(instances[0])
        metrics = check.get_metrics()

        # Make sure the base metrics loaded
        base_metrics = [m[0] for m in check.METRICS]
        ret_metrics = [m[0] for m in metrics]
        for metric in base_metrics:
            assert metric in ret_metrics

        # Check our custom metrics
        assert 'sqlserver.clr.execution' in ret_metrics
        assert 'sqlserver.exec.in_progress' in ret_metrics
        assert 'sqlserver.db.commit_table_entries' in ret_metrics

        # Make sure the ALL custom metric is tagged
        tagged_metrics = [m for m in metrics
                          if m[0] == 'sqlserver.db.commit_table_entries']
        for metric in tagged_metrics:
            for tag in metric[3]['dimensions']:
                assert tag.startswith('db')

if __name__ == "__main__":
    unittest.main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil
import unittest
import uuid

from common import get_check


class TestCrash(unittest.TestCase):

    def setUp(self):
        self.crash_dir = '/tmp/crash-test-%s' % str(uuid.uuid4())
        os.mkdir(self.crash_dir)

    def tearDown(self):
        shutil.rmtree(self.crash_dir)

    def test_checks(self):
        config = """
init_config:
   crash_dir: %s

instances:
   - name: crash_stats
""" % self.crash_dir

        (check, instances) = get_check('crash', config)

        # Baseline check
        check.check(instances[0])
        metrics = check.get_metrics()
        self.assertEqual(metrics[0].value, 0)
        self.assertEqual(metrics[0].value_meta['latest'], '')

        # Add a crash and re-check
        os.mkdir(os.path.join(self.crash_dir,'201504141011'))

        check.check(instances[0])
        metrics = check.get_metrics()
        self.assertEqual(metrics[0].value, 1)
        self.assertEqual(metrics[0].value_meta['latest'],
                         '2015-04-14 10:11:00')

        # Add a second crash and re-check
        os.mkdir(os.path.join(self.crash_dir,'201505222303'))

        check.check(instances[0])
        metrics = check.get_metrics()
        self.assertEqual(metrics[0].value, 2)
        self.assertEqual(metrics[0].value_meta['latest'],
                         '2015-05-22 23:03:00')


if __name__ == "__main__":
    unittest.main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from common import get_check
from random import shuffle, sample

import unittest
import os
import binascii
import re
import shutil
from nose.plugins.skip import SkipTest

class TestPostfix(unittest.TestCase):
    #
    # you can execute this dd-agent unit test via python's nose tool
    #
    # example:
    #
    #     nosetests --nocapture --tests=dd-agent/tests/test_postfix.py
    #

    def setUp(self):
        self.queue_root = '/tmp/dd-postfix-test/var/spool/postfix'

        self.queues = [
            'active',
            'maildrop',
            'bounce',
            'incoming',
            'deferred'
        ]

        self.in_count = {}

        # create test queues
        for queue in self.queues:
            try:
                os.makedirs(os.path.join(self.queue_root, queue))
                self.in_count[queue] = [0, 0]
            except Exception:
                pass

    def tearDown(self):
        # clean up test queues
        shutil.rmtree('/tmp/dd-postfix-test')

    def stripHeredoc(self, text):
        indent = len(min(re.findall('\n[ \t]*(?=\S)', text) or ['']))
        pattern = r'\n[ \t]{%d}' % (indent - 1)
        return re.sub(pattern, '\n', text)

    def test_checks(self):
        raise SkipTest('Requires root access to postfix')
        self.config = self.stripHeredoc("""init_config:

        instances:
            - directory: %s
              queues:
                  - bounce
                  - maildrop
                  - incoming
                  - active
                  - deferred
        """ % (self.queue_root))

        # stuff 10K msgs in random queues
        for _ in range(1, 10000):
            shuffle(self.queues)
            rand_queue = sample(self.queues, 1)[0]
            queue_file = binascii.b2a_hex(os.urandom(7))

            open(os.path.join(self.queue_root, rand_queue, queue_file), 'w')

            # keep track of what we put in
            self.in_count[rand_queue][0] += 1

        check, instances = get_check('postfix', self.config)

        check.check(instances[0])
        out_count = check.get_metrics()

        # output what went in... per queue
        print()
        for queue, count in self.in_count.items():
            print('Test messages put into', queue, '= ', self.in_count[queue][0])

        # output postfix.py dd-agent plugin counts... per queue
        print()
        for tuple in out_count:
            queue = tuple[3]['dimensions'][0].split(':')[1]
            self.assertEqual(int(tuple[2]), int(self.in_count[queue][0]))
            print('Test messages counted by dd-agent for', queue, '= ', tuple[2])

        #
        # uncomment this to see the raw dd-agent metric output
        #
        # print out_count

if __name__ == '__main__':
    unittest.main()



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import monasca_setup.detection


class Nova(monasca_setup.detection.ServicePlugin):

    """Detect Nova daemons and setup configuration to monitor them.

    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'compute',
            'process_names': ['nova-compute', 'nova-conductor',
                              'nova-cert', 'nova-network',
                              'nova-scheduler', 'nova-novncproxy',
                              'nova-xvpncproxy', 'nova-consoleauth',
                              'nova-objectstore', 'nova-api'],
            'service_api_url': 'http://localhost:8774/v2.0',
            'search_pattern': '.*version=2.*'
        }

        super(Nova, self).__init__(service_params)



# Copyright 2014 Intel
# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Implementation of Inspector abstraction for XenAPI."""

from eventlet import timeout
from oslo_config import cfg
from oslo_utils import units
try:
    import XenAPI as api
except ImportError:
    api = None

from monasca_agent.collector.virt import inspector as virt_inspector

opt_group = cfg.OptGroup(name='xenapi',
                         title='Options for XenAPI')

OPTS = [
    cfg.StrOpt('connection_url',
               help='URL for connection to XenServer/Xen Cloud Platform.'),
    cfg.StrOpt('connection_username',
               default='root',
               help='Username for connection to XenServer/Xen Cloud '
                    'Platform.'),
    cfg.StrOpt('connection_password',
               help='Password for connection to XenServer/Xen Cloud Platform.',
               secret=True),
    cfg.IntOpt('login_timeout',
               default=10,
               help='Timeout in seconds for XenAPI login.'),
]

CONF = cfg.CONF
CONF.register_group(opt_group)
CONF.register_opts(OPTS, group=opt_group)


def get_instance_name(instance):
    """Shortcut to get instance name."""
    return getattr(instance, 'OS-EXT-SRV-ATTR:instance_name', None)


class XenapiException(virt_inspector.InspectorException):
    pass


def get_api_session():
    if not api:
        raise ImportError('XenAPI not installed')

    url = CONF.xenapi.connection_url
    username = CONF.xenapi.connection_username
    password = CONF.xenapi.connection_password
    if not url or password is None:
        raise XenapiException('Must specify connection_url, and '
                              'connection_password to use')

    exception = api.Failure("Unable to log in to XenAPI "
                            "(is the Dom0 disk full?)")
    try:
        session = api.Session(url)
        with timeout.Timeout(CONF.xenapi.login_timeout, exception):
            session.login_with_password(username, password)
    except api.Failure as e:
        msg = "Could not connect to XenAPI: %s" % e.details[0]
        raise XenapiException(msg)
    return session


class XenapiInspector(virt_inspector.Inspector):

    def __init__(self):
        super(XenapiInspector, self).__init__()
        self.session = get_api_session()

    def _get_host_ref(self):
        """Return the xenapi host on which nova-compute runs on."""
        return self.session.xenapi.session.get_this_host(self.session.handle)

    def _call_xenapi(self, method, *args):
        return self.session.xenapi_request(method, args)

    def _lookup_by_name(self, instance_name):
        vm_refs = self._call_xenapi("VM.get_by_name_label", instance_name)
        n = len(vm_refs)
        if n == 0:
            raise virt_inspector.InstanceNotFoundException(
                'VM %s not found in XenServer' % instance_name)
        elif n > 1:
            raise XenapiException(
                'Multiple VM %s found in XenServer' % instance_name)
        else:
            return vm_refs[0]

    def inspect_cpu_util(self, instance, duration=None):
        instance_name = get_instance_name(instance)
        vm_ref = self._lookup_by_name(instance_name)
        metrics_ref = self._call_xenapi("VM.get_metrics", vm_ref)
        metrics_rec = self._call_xenapi("VM_metrics.get_record",
                                        metrics_ref)
        vcpus_number = metrics_rec['VCPUs_number']
        vcpus_utils = metrics_rec['VCPUs_utilisation']
        if len(vcpus_utils) == 0:
            msg = "Could not get VM %s CPU Utilization" % instance_name
            raise XenapiException(msg)

        utils = 0.0
        for num in range(int(vcpus_number)):
            utils += vcpus_utils.get(str(num))
        utils = utils / int(vcpus_number) * 100
        return virt_inspector.CPUUtilStats(util=utils)

    def inspect_memory_usage(self, instance, duration=None):
        instance_name = get_instance_name(instance)
        vm_ref = self._lookup_by_name(instance_name)
        metrics_ref = self._call_xenapi("VM.get_metrics", vm_ref)
        metrics_rec = self._call_xenapi("VM_metrics.get_record",
                                        metrics_ref)
        # Stat provided from XenServer is in B, converting it to MB.
        memory = int(metrics_rec['memory_actual']) / units.Mi
        return virt_inspector.MemoryUsageStats(usage=memory)

    def inspect_vnic_rates(self, instance, duration=None):
        instance_name = get_instance_name(instance)
        vm_ref = self._lookup_by_name(instance_name)
        vif_refs = self._call_xenapi("VM.get_VIFs", vm_ref)
        if vif_refs:
            for vif_ref in vif_refs:
                vif_rec = self._call_xenapi("VIF.get_record", vif_ref)
                vif_metrics_ref = self._call_xenapi(
                    "VIF.get_metrics", vif_ref)
                vif_metrics_rec = self._call_xenapi(
                    "VIF_metrics.get_record", vif_metrics_ref)

                interface = virt_inspector.Interface(
                    name=vif_rec['uuid'],
                    mac=vif_rec['MAC'],
                    fref=None,
                    parameters=None)
                rx_rate = float(vif_metrics_rec['io_read_kbs']) * units.Ki
                tx_rate = float(vif_metrics_rec['io_write_kbs']) * units.Ki
                stats = virt_inspector.InterfaceRateStats(rx_rate, tx_rate)
                yield (interface, stats)

    def inspect_disk_rates(self, instance, duration=None):
        instance_name = get_instance_name(instance)
        vm_ref = self._lookup_by_name(instance_name)
        vbd_refs = self._call_xenapi("VM.get_VBDs", vm_ref)
        if vbd_refs:
            for vbd_ref in vbd_refs:
                vbd_rec = self._call_xenapi("VBD.get_record", vbd_ref)
                vbd_metrics_ref = self._call_xenapi("VBD.get_metrics",
                                                    vbd_ref)
                vbd_metrics_rec = self._call_xenapi("VBD_metrics.get_record",
                                                    vbd_metrics_ref)

                disk = virt_inspector.Disk(device=vbd_rec['device'])
                # Stats provided from XenServer are in KB/s,
                # converting it to B/s.
                read_rate = float(vbd_metrics_rec['io_read_kbs']) * units.Ki
                write_rate = float(vbd_metrics_rec['io_write_kbs']) * units.Ki
                disk_rate_info = virt_inspector.DiskRateStats(
                    read_bytes_rate=read_rate,
                    read_requests_rate=0,
                    write_bytes_rate=write_rate,
                    write_requests_rate=0)
                yield(disk, disk_rate_info)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import logging


import monasca_agent.collector.checks as checks

try:
    import monasca_common.kafka_lib.client as client
    import monasca_common.kafka_lib.common as common
    import monasca_common.kafka_lib.consumer as consumer
except ImportError:
    raise Exception('Missing python dependency: monasca-common.\n Please '
                    'install with: pip install monasca-agent[kafka_plugin]')

log = logging.getLogger(__name__)


class KafkaConnection(object):
    """A simple context manager for kafka connections """

    def __init__(self, connect_str):
        self.connect_str = connect_str

    def __enter__(self):
        self.kafka_conn = client.KafkaClient(self.connect_str)
        return self.kafka_conn

    def __exit__(self, type, value, traceback):
        try:
            self.kafka_conn.close()
        except Exception:
            log.exception('Error cleaning up Kafka connection')


class KafkaCheck(checks.AgentCheck):
    """Checks the configured kafka instance reporting the consumption lag
       for each partition per topic in each consumer group. If full_output
       is set also reports broker offsets and the current consumer offset.
       Works on Kafka version >= 0.8.1.1
    """
    def _parse_consumer_groups(self, raw_val):
        """Parses and validates the config
        Expected format is:
        consumer_groups:
          myconsumer0: # consumer group name
            - mytopic0
          myconsumer1:
            - mytopic0
            - mytopic1
        """
        consumer_groups = dict()

        try:
            for group, topics in raw_val.items():
                assert isinstance(group, str)
                if isinstance(topics, dict):
                    self.log.info("Found old config format, discarding partition list")
                    topics = list(topics.keys())
                assert isinstance(topics[0], str)
                consumer_groups[group] = topics
            return consumer_groups
        except Exception as e:
            self.log.exception(e)
            raise Exception("Invalid `consumer_groups` value. Must be a mapping of lists")

    def _get_kafka_offsets(self, kafka_conn, consumer_groups):
        # Query Kafka for consumer offsets
        consumer_offsets = {}
        topic_partitions = collections.defaultdict(set)
        for consumer_group, topics in consumer_groups.items():
            for topic in topics:
                kafka_consumer = None
                try:
                    kafka_consumer = consumer.SimpleConsumer(kafka_conn,
                                                             consumer_group,
                                                             topic,
                                                             auto_commit=False)
                    kafka_consumer.fetch_last_known_offsets()

                    partitions = kafka_consumer.offsets.keys()
                except Exception:
                    self.log.error('Error fetching partition list for topic {0}'.format(topic))
                    if kafka_consumer is not None:
                        kafka_consumer.stop()
                    continue

                # Remember the topic partitions encountered so that we can look up their
                # broker offsets later
                topic_partitions[topic].update(set(partitions))
                consumer_offsets[(consumer_group, topic)] = {}
                for partition in partitions:
                    try:
                        consumer_offsets[(consumer_group,
                                          topic)][partition] = kafka_consumer.offsets[partition]
                    except KeyError:
                        self.log.error(
                            'Error fetching consumer offset for {0} partition {1}'.format(
                                topic, partition))

                kafka_consumer.stop()

        # Query Kafka for the broker offsets, done in a separate loop so only one query is done
        # per topic/partition even if multiple consumer groups watch the same topic
        broker_offsets = {}
        for topic, partitions in topic_partitions.items():
            offset_responses = []
            for p in partitions:
                try:
                    response = kafka_conn.send_offset_request(
                        [common.OffsetRequest(topic.encode('utf-8'), p, -1, 1)])
                    offset_responses.append(response[0])
                except common.KafkaError as e:
                    self.log.error("Error fetching broker offset: {0}".format(e))
            for resp in offset_responses:
                broker_offsets[(resp.topic, resp.partition)] = resp.offsets[0]
        return consumer_offsets, broker_offsets

    def check(self, instance):
        raw_consumer_groups = self.read_config(instance, 'consumer_groups')
        consumer_groups = self._parse_consumer_groups(raw_consumer_groups)

        kafka_host_ports = self.read_config(instance, 'kafka_connect_str')
        full_output = self.read_config(instance, 'full_output', cast=bool, optional=True)
        per_partition = self.read_config(instance, 'per_partition', cast=bool, optional=True)
        if not per_partition:
            full_output = False
        dimensions = {'component': 'kafka', 'service': 'kafka'}

        # Connect to Kafka and pull information
        with KafkaConnection(kafka_host_ports) as kafka_conn:
            consumer_offsets, broker_offsets = self._get_kafka_offsets(kafka_conn, consumer_groups)
        # Report the broker data if full output
        if full_output:
            broker_dimensions = dimensions.copy()
            for (topic, partition), broker_offset in broker_offsets.items():
                broker_dimensions.update({'topic': topic, 'partition': str(partition)})
                broker_offset = broker_offsets.get((topic.encode('utf-8'), partition))
                self.gauge('kafka.broker_offset', broker_offset,
                           dimensions=self._set_dimensions(broker_dimensions, instance))

        # Report the consumer data
        consumer_dimensions = dimensions.copy()
        for (consumer_group, topic), offsets in consumer_offsets.items():
            if per_partition:
                for partition, consumer_offset in offsets.items():
                    # Get the broker offset
                    broker_offset = broker_offsets.get((topic.encode('utf-8'), partition))
                    # Report the consumer offset and lag
                    consumer_dimensions.update({'topic': topic, 'partition': str(partition),
                                                'consumer_group': consumer_group})
                    if full_output:
                        self.gauge('kafka.consumer_offset', consumer_offset,
                                   dimensions=self._set_dimensions(consumer_dimensions, instance))
                    self.gauge('kafka.consumer_lag', broker_offset - consumer_offset,
                               dimensions=self._set_dimensions(consumer_dimensions, instance))
            else:
                consumer_dimensions.update({'topic': topic, 'consumer_group': consumer_group})
                total_lag = 0
                for partition, consumer_offset in offsets.items():
                    broker_offset = broker_offsets.get((topic.encode('utf-8'), partition))
                    total_lag += broker_offset - consumer_offset

                self.gauge('kafka.consumer_lag', total_lag,
                           dimensions=self._set_dimensions(consumer_dimensions, instance))
